import datetime as dt
import json
import logging
import os

import dask
import numpy as np
import pandas as pd
import xarray as xr

from icenet.data.producers import Processor
from icenet.data.sic.mask import Masks
from icenet.model.models import linear_trend_forecast

"""

"""


class IceNetPreProcessor(Processor):
    """

    :param abs_vars: 
    :param anom_vars: 
    :param name: 
    :param train_dates:
    :param val_dates: 
    :param test_dates: 
    :param *args: 
    :param data_shape: 
    :param dtype: 
    :param exclude_vars: 
    :param file_filters: 
    :param identifier: 
    :param linear_trends: 
    :param linear_trend_days: 
    :param meta_vars: 
    :param missing_dates: 
    :param minmax: 
    :param no_normalise: 
    :param path: 
    :param ref_procdir: 
    :param source_data: 
    :param update_key: 
    :param update_loader: 
    """

    DATE_FORMAT = "%Y_%m_%d"

    def __init__(self,
                 abs_vars,
                 anom_vars,
                 name,
                 # FIXME: the preprocessors don't need to have the concept of
                 #  train, test, val: they only need to output daily files
                 #  that either are, or are not, part of normalisation /
                 #  climatology calculations. Not a problem, just fix
                 train_dates,
                 val_dates,
                 test_dates,
                 *args,
                 data_shape=(432, 432),
                 dtype=np.float32,
                 exclude_vars=(),
                 file_filters=tuple(["latlon_"]),
                 identifier=None,
                 linear_trends=tuple(["siconca"]),
                 linear_trend_steps=7,
                 meta_vars=tuple(),
                 missing_dates=tuple(),
                 minmax=True,
                 no_normalise=tuple(["siconca"]),
                 path=os.path.join(".", "processed"),
                 ref_procdir=None,
                 source_data=os.path.join(".", "data"),
                 update_key=None,
                 update_loader=True,
                 **kwargs):
        super().__init__(identifier,
                         source_data,
                         *args,
                         file_filters=file_filters,
                         path=os.path.join(path, name),
                         train_dates=train_dates,
                         val_dates=val_dates,
                         test_dates=test_dates,
                         **kwargs)

        self._abs_vars = abs_vars if abs_vars else []
        self._anom_vars = anom_vars if anom_vars else []
        # TODO: Ugh, this should not be here any longer
        self._meta_vars = list(meta_vars)

        self._name = name

        self._data_shape = data_shape
        self._dtype = dtype
        self._exclude_vars = exclude_vars
        self._linear_trends = linear_trends
        self._missing_dates = list(missing_dates)
        self._no_normalise = no_normalise
        self._normalise = self._normalise_array_mean \
            if not minmax else self._normalise_array_scaling
        self._refdir = ref_procdir
        self._update_key = self.identifier if not update_key else update_key
        self._update_loader = os.path.join(".",
                                           "loader.{}.json".format(name)) \
            if update_loader else None

        if type(linear_trend_steps) == int:
            logging.debug("Setting range for linear trend steps based on {}".
                          format(linear_trend_steps))
            self._linear_trend_steps = list(range(1, linear_trend_steps + 1))
        else:
            self._linear_trend_steps = [int(el) for el in linear_trend_steps]

    def process(self):
        """

        """
        var_suffixes = ["abs", "anom"]
        var_lists = [self._abs_vars, self._anom_vars]
        for var_suffix, var_list in zip(var_suffixes, var_lists):
            for var_name in var_list:
                if var_name not in self._var_files.keys():
                    logging.warning("{} does not exist".format(var_name))
                    continue
                self._save_variable(var_name, var_suffix)

        if self._update_loader:
            self.update_loader_config()

    def pre_normalisation(self, var_name: str, da: object):
        """

        :param var_name:
        :param da:
        :return:
        """
        logging.debug("No pre normalisation implemented for {}".
                      format(var_name))
        return da

    def post_normalisation(self, var_name: str, da: object):
        """

        :param var_name:
        :param da:
        :return:
        """
        logging.debug("No post normalisation implemented for {}".
                      format(var_name))
        return da

    # TODO: update this to store parameters, if appropriate
    def update_loader_config(self):
        """

        :return:
        """
        def _serialize(x):
            if x is dt.date:
                return x.strftime(IceNetPreProcessor.DATE_FORMAT)
            return str(x)

        # We have to be explicit with "dates" as the properties will not be
        # caught by _serialize
        source = {
            "name":             self._name,
            "implementation":   self.__class__.__name__,
            "anom":             self._anom_vars,
            "abs":              self._abs_vars,
            "dates":            {
                "train":        [d.strftime(IceNetPreProcessor.DATE_FORMAT)
                                 for d in self._dates.train],
                "val":          [d.strftime(IceNetPreProcessor.DATE_FORMAT)
                                 for d in self._dates.val],
                "test":         [d.strftime(IceNetPreProcessor.DATE_FORMAT)
                                 for d in self._dates.test],
            },
            "linear_trends":    self._linear_trends,
            "linear_trend_steps": self._linear_trend_steps,
            "meta":             self._meta_vars,
            # TODO: intention should perhaps be to strip these from
            #  other date sets, this is just an indicative placeholder
            #  for the mo
            "var_files":        self._processed_files,
        }

        configuration = {
            "sources": {}
        }

        if os.path.exists(self._update_loader):
            logging.info("Loading configuration {}".format(self._update_loader))
            with open(self._update_loader, "r") as fh:
                obj = json.load(fh)
                configuration.update(obj)

        configuration["sources"][self._update_key] = source

        # Ideally should always be in together
        if "dtype" in configuration:
            assert configuration["dtype"] == self._dtype.__name__

        if "shape" in configuration:
            assert configuration["shape"] == list(self._data_shape)

        configuration["dtype"] = self._dtype.__name__
        configuration["shape"] = list(self._data_shape)

        if "missing_dates" not in configuration:
            configuration["missing_dates"] = []

        # Conversion required one way or another, so perhaps more efficient
        # than a union
        for d in sorted(self._missing_dates):
            date_str = d.strftime(IceNetPreProcessor.DATE_FORMAT)

            if date_str not in configuration["missing_dates"]:
                configuration["missing_dates"].append(date_str)

        logging.info("Writing configuration to {}".format(self._update_loader))

        with open(self._update_loader, "w") as fh:
            json.dump(configuration, fh, indent=4, default=_serialize)

    def _save_variable(self, var_name: str, var_suffix: str):
        """

        :param var_name:
        :param var_suffix:
        """
        with dask.config.set(**{'array.slicing.split_large_chunks': True}):
            da = self._open_dataarray_from_files(var_name)

            # FIXME: we should ideally store train dates against the
            #  normalisation and climatology, to ensure recalculation on
            #  reprocess. All this need be is in the path, to be honest

            if var_name in self._anom_vars:
                if self._refdir:
                    logging.info("Loading climatology from alternate "
                                 "directory: {}".format(self._refdir))
                    clim_path = os.path.join(self._refdir,
                                             "params",
                                             "climatology.{}".format(var_name))
                else:
                    clim_path = os.path.join(self.get_data_var_folder("params"),
                                             "climatology.{}".format(var_name))

                if not os.path.exists(clim_path):
                    logging.info("Generating climatology {}".format(clim_path))

                    if self._dates.train:
                        climatology = da.sel(time=self._dates.train).\
                            groupby('time.month', restore_coord_dims=True).\
                            mean()

                        climatology.to_netcdf(clim_path)
                    else:
                        raise RuntimeError("{} does not exist and no "
                                           "training data is supplied".
                                           format(clim_path))
                else:
                    logging.info("Reusing climatology {}".format(clim_path))
                    climatology = xr.open_dataarray(clim_path)

                if not set(da.groupby("time.month").all().month.values).\
                        issubset(set(climatology.month.values)):
                    logging.warning(
                        "We don't have a full climatology ({}) "
                        "compared with data ({})".format(
                            ",".join([str(i)
                                      for i in climatology.month.values]),
                            ",".join([str(i)
                                      for i in da.groupby("time.month").
                                     all().month.values])))
                    da = da - climatology.mean()
                else:
                    da = da.groupby("time.month") - climatology

            # FIXME: this is not the way to reconvert underlying data on
            #  dask arrays
            da.data = np.asarray(da.data, dtype=self._dtype)

            da = self.pre_normalisation(var_name, da)
            # We don't do this (https://github.com/tom-andersson/icenet2/blob/
            # 4ca0f1300fbd82335d8bb000c85b1e71855630fa/icenet/utils.py#L520)
            # any more

            if var_name in self._linear_trends and var_suffix == "abs":
                # TODO: verify, this used to be da = , but we should not be
                #  overwriting the abs da with linear trend da
                ref_da = None

                if self._refdir:
                    logging.info("We have a reference {}, so will load "
                                 "and supply abs from that for linear trend of "
                                 "{}".format(self._refdir, var_name))
                    ref_da = xr.open_dataarray(
                        os.path.join(self._refdir, var_name,
                                     "{}_{}.nc".format(var_name, var_suffix)))

                self._build_linear_trend_da(da, var_name, ref_da=ref_da)

            elif var_name in self._linear_trends \
                    and var_name not in self._abs_vars:
                raise NotImplementedError("You've asked for linear trend "
                                          "without an  absolute value var: {}".
                                          format(var_name))

            if var_name in self._no_normalise:
                logging.info("No normalisation for {}".format(var_name))
            else:
                logging.info("Normalising {}".format(var_name))
                da = self._normalise(var_name, da)

            da = self.post_normalisation(var_name, da)

            self.save_processed_file(var_name,
                                     "{}_{}.nc".format(var_name, var_suffix),
                                     da.rename(
                                         "_".join([var_name, var_suffix])))

    def _open_dataarray_from_files(self, var_name: str):

        """
        Open the yearly xarray files, accounting for some ERA5 variables that
        have erroneous 'unknown' NetCDF variable names which prevents
        concatentation.

        :param var_name:
        :return:
        """

        logging.info("Opening files for {}".format(var_name))
        logging.debug("Files: {}".format(self._var_files[var_name]))
        ds = xr.open_mfdataset(self._var_files[var_name],
                               # Solves issue with inheriting files without
                               # time dimension (only having coordinate)
                               combine="nested",
                               concat_dim="time",
                               coords="minimal",
                               compat="override",
                               drop_variables=("lat", "lon"),
                               # TODO: Wasteful on small sets, but much faster
                               #  on big sets: make optional
                               parallel=True)

        # For processing one file, we're going to assume a single non-lambert
        # variable exists at the start and rename all of them
        var_names = [name for name in list(ds.data_vars.keys())
                     if not name.startswith("lambert_")]

        var_names = set(var_names)
        logging.debug("Files have var names {} which will be renamed to {}".
                      format(", ".join(var_names), var_name))

        ds = ds.rename({k: var_name for k in var_names})
        da = getattr(ds, var_name)

        # all_dates = self.dates.train + self.dates.val + self.dates.test
        # logging.debug("{} dates in total".format(len(all_dates)))

        da_dates = [pd.to_datetime(d).date() for d in da.time.values]
        logging.debug("{} dates in da".format(len(da_dates)))

        # search = sorted(list(set([el for el in all_dates
        #                          if pd.to_datetime(el).date() in da_dates])))
        # logging.debug("Selecting {} dates from da".format(len(search)))

        # We no longer select on all_dates, as it destroys lag/lead processed
        # selections from the dataset
        try:
            da = da.sel(time=da_dates)
        except KeyError:
            # There is likely non-resampled data being used
            # TODO: we could use nearest neighbour on this coordinate,
            #  but this feels more reliable to dodgy input data when
            #  transferring
            logging.warning("Data selection failed, likely not daily sampled "
                            "data so will give that a try")
            da = da.resample(time="1D").mean().sel(time=da_dates).sortby("time")
        logging.info("Filtered to {} units long based on configuration "
                     "requirements".format(len(da.time)))

        for coord in ["yc", "xc"]:
            if getattr(da, coord).attrs['units'] == "km":
                da[coord] = da[coord] * 1000
                da[coord].attrs['units'] = "meters"
        return da

    @staticmethod
    def mean_and_std(array: object):
        """
        Return the mean and standard deviation of an array-like object (intended
        use case is for normalising a raw satellite data array based on a list
        of samples used for training).
        :param array:
        :return:
        """

        mean = np.nanmean(array)
        std = np.nanstd(array)

        logging.info("Mean: {:.3f}, std: {:.3f}".
                     format(mean.item(), std.item()))

        return mean, std

    def _normalise_array_mean(self, var_name: str, da: object):

        """
        Using the *training* data only, compute the mean and
        standard deviation of the input raw satellite DataArray (`da`)
        and return a normalised version. If minmax=True,
        instead normalise to lie between min and max of the elements of `array`.

        If min, max, mean, or std are given values other than None,
        those values are used rather than being computed from the training
        months.

        :param var_name:
        :param da:
        :return:
        """

        if self._refdir:
            logging.info("Using alternate processing directory {} for "
                         "mean".format(self._refdir))
            proc_dir = os.path.join(self._refdir, "normalisation.mean")
        else:
            proc_dir = self.get_data_var_folder("normalisation.mean")

        mean_path = os.path.join(proc_dir, "{}".format(var_name))

        if os.path.exists(mean_path):
            logging.debug("Loading norm-average mean-std from {}".
                          format(mean_path))
            mean, std = tuple([self._dtype(el) for el in
                               open(mean_path, "r").read().split(",")])
        elif self._dates.train:
            logging.debug("Generating norm-average mean-std from {} training "
                          "dates".format(len(self._dates.train)))
            training_samples = da.sel(time=self._dates.train).data
            training_samples = training_samples.ravel()

            mean, std = IceNetPreProcessor.mean_and_std(training_samples)
        else:
            raise RuntimeError("Either a normalisation file or training data "
                               "must be supplied")

        new_da = (da - mean) / std

        if not self._refdir:
            open(mean_path, "w").write(",".join([str(f) for f in
                                                [mean, std]]))
        return new_da

    def _normalise_array_scaling(self, var_name: str, da: object):
        """

        :param var_name:
        :param da:
        :return:
        """
        if self._refdir:
            logging.info("Using alternate processing directory {} for "
                         "scaling".format(self._refdir))
            proc_dir = os.path.join(self._refdir, "normalisation.scale")
        else:
            proc_dir = self.get_data_var_folder("normalisation.scale")

        scale_path = os.path.join(proc_dir, "{}".format(var_name))

        if os.path.exists(scale_path):
            logging.debug("Loading norm-scaling min-max from {}".
                          format(scale_path))
            minimum, maximum = tuple([self._dtype(el) for el in
                               open(scale_path, "r").read().split(",")])
        elif self._dates.train:
            logging.debug("Generating norm-scaling min-max from {} training "
                          "dates".format(len(self._dates.train)))
            training_samples = da.sel(time=self._dates.train).data
            training_samples = training_samples.ravel()

            minimum = np.nanmin(training_samples).astype(self._dtype)
            maximum = np.nanmax(training_samples).astype(self._dtype)
        else:
            raise RuntimeError("Either a normalisation file or training data "
                               "must be supplied")

        new_da = (da - minimum) / (maximum - minimum)
        if not self._refdir:
            open(scale_path, "w").write(",".join([str(f) for f in
                                                 [minimum, maximum]]))
        return new_da

    def _build_linear_trend_da(self,
                               input_da: object,
                               var_name: str,
                               max_years: int = 35,
                               ref_da: object = None):
        """
        Construct a DataArray `linear_trend_da` containing the linear trend
        forecasts based on the input DataArray `input_da`.

        :param input_da:
        :param var_name:
        :param max_years:
        :param ref_da:
        :return:
        """

        if ref_da is None:
            ref_da = input_da
        data_dates = sorted([pd.Timestamp(date)
                             for date in input_da.time.values])

        trend_dates = set()
        trend_steps = max(self._linear_trend_steps)
        logging.info("Generating trend data up to {} steps ahead for {} dates".
                     format(trend_steps, len(data_dates)))

        for dat_date in data_dates:
            trend_dates = trend_dates.union(
                [dat_date + pd.DateOffset(days=d)
                 for d in self._linear_trend_steps])

        trend_dates = list(sorted(trend_dates))
        logging.info("Generating {} trend dates".format(len(trend_dates)))

        linear_trend_da = \
            xr.broadcast(input_da, xr.DataArray(pd.date_range(
                data_dates[0],
                data_dates[-1] + pd.DateOffset(days=trend_steps)),
                    dims="time"))[0]
        linear_trend_da = linear_trend_da.sel(time=trend_dates)
        linear_trend_da.data = np.zeros(linear_trend_da.shape)

        land_mask = Masks(north=self.north, south=self.south).get_land_mask()

        # Could use shelve, but more likely we'll run into concurrency issues
        # pickleshare might be an option but a little over-engineery
        trend_cache_path = os.path.join(
            self.get_data_var_folder(var_name),
            "{}_linear_trend.nc".format(var_name)
        )
        trend_cache = linear_trend_da.copy()
        trend_cache.data = np.full_like(linear_trend_da.data, np.nan)

        if os.path.exists(trend_cache_path):
            trend_cache = xr.open_dataarray(trend_cache_path)
            logging.info("Loaded {} entries from {}".
                         format(len(trend_cache.time), trend_cache_path))

        def data_selector(da,
                          processing_date,
                          missing_dates=tuple()):
            target_date = pd.to_datetime(processing_date)

            date_da = da[(da.time['time.month'] == target_date.month) &
                         (da.time['time.day'] == target_date.day) &
                         (da.time <= target_date) &
                         ~da.time.isin(missing_dates)].\
                isel(time=slice(0, max_years))
            return date_da

        for forecast_date in sorted(trend_dates, reverse=True):
            if not trend_cache.sel(time=forecast_date).isnull().all():
                output_map = trend_cache.sel(time=forecast_date)
            else:
                output_map = linear_trend_forecast(
                    data_selector, forecast_date, ref_da, land_mask,
                    missing_dates=self._missing_dates,
                    shape=self._data_shape)

            linear_trend_da.loc[dict(time=forecast_date)] = output_map

        logging.info("Writing new trend cache for {}".format(var_name))
        trend_cache.close()
        linear_trend_da = linear_trend_da.rename(
            "{}_linear_trend".format(var_name))
        self.save_processed_file(var_name,
                                 "{}_linear_trend.nc".format(var_name),
                                 linear_trend_da)

        return linear_trend_da

    @property
    def missing_dates(self):
        return self._missing_dates

    @missing_dates.setter
    def missing_dates(self, arr):
        self._missing_dates = arr



import copy
import fnmatch
import ftplib
import logging
import os

import datetime as dt
from ftplib import FTP

import numpy as np
import pandas as pd
import xarray as xr

from icenet.data.cli import download_args
from icenet.data.producers import Downloader
from icenet.data.sic.mask import Masks
from icenet.utils import Hemisphere
from icenet.data.sic.utils import SIC_HEMI_STR

"""

"""

invalid_sic_days = {
    Hemisphere.NORTH: [
        *[d.date() for d in
          pd.date_range(dt.date(1979, 5, 21), dt.date(1979, 6, 4))],
        *[d.date() for d in
          pd.date_range(dt.date(1979, 6, 10), dt.date(1979, 6, 26))],
        dt.date(1979, 7, 1),
        *[d.date() for d in
          pd.date_range(dt.date(1979, 7, 24), dt.date(1979, 7, 28))],
        *[d.date() for d in
          pd.date_range(dt.date(1980, 1, 4), dt.date(1980, 1, 10))],
        *[d.date() for d in
          pd.date_range(dt.date(1980, 2, 27), dt.date(1980, 3, 4))],
        *[d.date() for d in
          pd.date_range(dt.date(1980, 3, 16), dt.date(1980, 3, 22))],
        *[d.date() for d in
          pd.date_range(dt.date(1980, 4, 9), dt.date(1980, 4, 15))],
        *[d.date() for d in
          pd.date_range(dt.date(1981, 2, 27), dt.date(1981, 3, 5))],
        *[d.date() for d in
          pd.date_range(dt.date(1984, 8, 12), dt.date(1984, 8, 24))],
        dt.date(1984, 9, 14),
        *[d.date() for d in
          pd.date_range(dt.date(1985, 9, 22), dt.date(1985, 9, 28))],
        *[d.date() for d in
          pd.date_range(dt.date(1986, 3, 29), dt.date(1986, 7, 1))],
        *[d.date() for d in
          pd.date_range(dt.date(1987, 1, 3), dt.date(1987, 1, 19))],
        *[d.date() for d in
          pd.date_range(dt.date(1987, 1, 29), dt.date(1987, 2, 2))],
        dt.date(1987, 2, 23),
        *[d.date() for d in
          pd.date_range(dt.date(1987, 2, 26), dt.date(1987, 3, 2))],
        dt.date(1987, 3, 13),
        *[d.date() for d in
          pd.date_range(dt.date(1987, 3, 22), dt.date(1987, 3, 26))],
        *[d.date() for d in
          pd.date_range(dt.date(1987, 4, 3), dt.date(1987, 4, 17))],
        *[d.date() for d in
          pd.date_range(dt.date(1987, 12, 1), dt.date(1988, 1, 12))],
        dt.date(1989, 1, 3),
        *[d.date() for d in
          pd.date_range(dt.date(1990, 12, 21), dt.date(1990, 12, 26))],
        dt.date(1979, 5, 28),
        dt.date(1979, 5, 30),
        dt.date(1979, 6, 1),
        dt.date(1979, 6, 3),
        dt.date(1979, 6, 11),
        dt.date(1979, 6, 13),
        dt.date(1979, 6, 15),
        dt.date(1979, 6, 17),
        dt.date(1979, 6, 19),
        dt.date(1979, 6, 21),
        dt.date(1979, 6, 23),
        dt.date(1979, 6, 25),
        dt.date(1979, 7, 1),
        dt.date(1979, 7, 25),
        dt.date(1979, 7, 27),
        dt.date(1984, 9, 14),
        dt.date(1987, 1, 16),
        dt.date(1987, 1, 18),
        dt.date(1987, 1, 30),
        dt.date(1987, 2, 1),
        dt.date(1987, 2, 23),
        dt.date(1987, 2, 27),
        dt.date(1987, 3, 1),
        dt.date(1987, 3, 13),
        dt.date(1987, 3, 23),
        dt.date(1987, 3, 25),
        dt.date(1987, 4, 4),
        dt.date(1987, 4, 6),
        dt.date(1987, 4, 10),
        dt.date(1987, 4, 12),
        dt.date(1987, 4, 14),
        dt.date(1987, 4, 16),
        dt.date(1987, 4, 4),
        dt.date(1990, 1, 26)
    ],
    Hemisphere.SOUTH: [
        dt.date(1979, 2, 5),
        dt.date(1979, 2, 25),
        dt.date(1979, 3, 23),
        *[d.date() for d in
          pd.date_range(dt.date(1979, 3, 26), dt.date(1979, 3, 30))],
        dt.date(1979, 4, 12),
        dt.date(1979, 5, 16),
        *[d.date() for d in
          pd.date_range(dt.date(1979, 5, 21), dt.date(1979, 5, 27))],
        *[d.date() for d in
          pd.date_range(dt.date(1979, 7, 10), dt.date(1979, 7, 18))],
        dt.date(1979, 8, 10),
        dt.date(1979, 9, 3),
        *[d.date() for d in
          pd.date_range(dt.date(1980, 1, 4), dt.date(1980, 1, 10))],
        dt.date(1980, 2, 16),
        *[d.date() for d in
          pd.date_range(dt.date(1980, 2, 27), dt.date(1980, 3, 4))],
        *[d.date() for d in
          pd.date_range(dt.date(1980, 3, 14), dt.date(1980, 3, 22))],
        dt.date(1980, 3, 31),
        *[d.date() for d in
          pd.date_range(dt.date(1980, 4, 9), dt.date(1980, 4, 15))],
        dt.date(1980, 4, 22),
        *[d.date() for d in
          pd.date_range(dt.date(1981, 2, 27), dt.date(1981, 3, 5))],
        dt.date(1981, 6, 10),
        *[d.date() for d in
          pd.date_range(dt.date(1981, 8, 3), dt.date(1982, 8, 9))],
        dt.date(1982, 8, 6),
        *[d.date() for d in
          pd.date_range(dt.date(1983, 7, 7), dt.date(1983, 7, 11))],
        dt.date(1983, 7, 22),
        dt.date(1984, 6, 12),
        *[d.date() for d in
          pd.date_range(dt.date(1984, 8, 12), dt.date(1984, 8, 24))],
        *[d.date() for d in
          pd.date_range(dt.date(1984, 9, 13), dt.date(1984, 9, 17))],
        *[d.date() for d in
          pd.date_range(dt.date(1984, 10, 3), dt.date(1984, 10, 9))],
        *[d.date() for d in
          pd.date_range(dt.date(1984, 11, 18), dt.date(1984, 11, 22))],
        dt.date(1985, 7, 23),
        *[d.date() for d in
          pd.date_range(dt.date(1985, 9, 22), dt.date(1985, 9, 28))],
        *[d.date() for d in
          pd.date_range(dt.date(1986, 3, 29), dt.date(1986, 11, 2))],
        *[d.date() for d in
          pd.date_range(dt.date(1987, 1, 3), dt.date(1987, 1, 15))],
        *[d.date() for d in
          pd.date_range(dt.date(1987, 12, 1), dt.date(1988, 1, 12))],
        dt.date(1990, 8, 14),
        dt.date(1990, 8, 15),
        dt.date(1990, 8, 24),
        *[d.date() for d in
          pd.date_range(dt.date(1990, 12, 22), dt.date(1990, 12, 26))],
        dt.date(1979, 2, 5),
        dt.date(1979, 2, 25),
        dt.date(1979, 3, 23),
        dt.date(1979, 3, 27),
        dt.date(1979, 3, 29),
        dt.date(1979, 4, 12),
        dt.date(1979, 5, 16),
        dt.date(1979, 7, 11),
        dt.date(1979, 7, 13),
        dt.date(1979, 7, 15),
        dt.date(1979, 7, 17),
        dt.date(1979, 8, 10),
        dt.date(1979, 9, 3),
        dt.date(1980, 2, 16),
        dt.date(1980, 3, 15),
        dt.date(1980, 3, 31),
        dt.date(1980, 4, 22),
        dt.date(1981, 6, 10),
        dt.date(1982, 8, 6),
        dt.date(1983, 7, 8),
        dt.date(1983, 7, 10),
        dt.date(1983, 7, 22),
        dt.date(1984, 6, 12),
        dt.date(1984, 9, 14),
        dt.date(1984, 9, 16),
        dt.date(1984, 10, 4),
        dt.date(1984, 10, 6),
        dt.date(1984, 10, 8),
        dt.date(1984, 11, 19),
        dt.date(1984, 11, 21),
        dt.date(1985, 7, 23),
        *pd.date_range(dt.date(1986, 7, 2), dt.date(1986, 11, 1)),
        dt.date(1990, 8, 14),
        dt.date(1990, 8, 15),
        dt.date(1990, 8, 24)
    ]
}

var_remove_list = ['time_bnds', 'raw_ice_conc_values', 'total_standard_error',
                   'smearing_standard_error', 'algorithm_standard_error',
                   'status_flag', 'Lambert_Azimuthal_Grid']


class SICDownloader(Downloader):
    """Downloads OSI-SAF SIC data from 1979-present using OpenDAP.

    The dataset comprises OSI-450 (1979-2015) and OSI-430-b (2016-ownards)
    Monthly averages are-computed on the server-side.
    This script can take about an hour to run.

    The query URLs were obtained from the following sites:
        - OSI-450 (1979-2016): https://thredds.met.no/thredds/dodsC/osisaf/
            met.no/reprocessed/ice/conc_v2p0_nh_agg.html
        - OSI-430-b (2016-present): https://thredds.met.no/thredds/dodsC/osisaf/
            met.no/reprocessed/ice/conc_crb_nh_agg.html

    :param additional_invalid_dates:
    :param dates:
    :param delete_tempfiles:
    :param download:
    :param dtype:
    """
    def __init__(self,
                 *args,
                 additional_invalid_dates: object = (),
                 dates: object = (),
                 delete_tempfiles: bool = True,
                 download: bool = True,
                 dtype: object = np.float32,
                 **kwargs):
        super().__init__(*args, identifier="osisaf", **kwargs)

        self._dates = dates
        self._delete = delete_tempfiles
        self._download = download
        self._dtype=dtype
        self._invalid_dates = invalid_sic_days[self.hemisphere] + \
            list(additional_invalid_dates)
        self._masks = Masks(north=self.north, south=self.south)

        self._mask_dict = {
            month: self._masks.get_active_cell_mask(month)
            for month in np.arange(1, 12+1)
        }

    def download(self):
        """

        """
        hs = SIC_HEMI_STR[self.hemisphere_str[0]]

        logging.info(
            "Not downloading SIC files, (re)processing NC files in "
            "existence already" if not self._download else
            "Downloading SIC datafiles to .temp intermediates...")

        ftp_osi450 = "/reprocessed/ice/conc/v2p0/{:04d}/{:02d}/"
        ftp_osi430b = "/reprocessed/ice/conc-cont-reproc/v2p0/{:04d}/{:02d}/"
        cache = {}
        osi430b_start = dt.date(2016, 1, 1)

        # TODO: filter based on existing data
        dt_arr = list(reversed(sorted(copy.copy(self._dates))))
        data_files = []
        ftp = None
        var = "siconca"

        while len(dt_arr):
            el = dt_arr.pop()

            if el in self._invalid_dates:
                logging.warning("Date {} is in invalid list".format(el))
                continue

            date_str = el.strftime("%Y_%m_%d")
            temp_path = os.path.join(
                self.get_data_var_folder(var, append=[str(el.year)]),
                "{}.temp".format(date_str))
            nc_path = os.path.join(
                self.get_data_var_folder(var, append=[str(el.year)]),
                "{}.nc".format(date_str))

            if not self._download:
                if os.path.exists(nc_path):
                    reproc_path = os.path.join(
                        self.get_data_var_folder(var,
                                                 append=[str(el.year)]),
                        "{}.reproc.nc".format(date_str))

                    logging.debug("{} exists, becoming {}".
                                  format(nc_path, reproc_path))
                    os.rename(nc_path, reproc_path)
                    data_files.append(reproc_path)
                else:
                    if os.path.exists(temp_path):
                        logging.info("Using existing {}".format(temp_path))
                        data_files.append(temp_path)
                    else:
                        logging.debug("{} does not exist".format(nc_path))
                continue
            else:
                if not os.path.isdir(os.path.dirname(temp_path)):
                    os.makedirs(os.path.dirname(temp_path), exist_ok=True)

                if os.path.exists(temp_path) or os.path.exists(nc_path):
                    logging.debug("{} file exists, skipping".format(date_str))
                    if not os.path.exists(nc_path):
                        data_files.append(temp_path)
                    continue

                if not ftp:
                    logging.info("FTP opening")
                    ftp = FTP('osisaf.met.no')
                    ftp.login()

                chdir_path = ftp_osi450 if el < osi430b_start else ftp_osi430b
                chdir_path = chdir_path.format(el.year, el.month)

                try:
                    ftp.cwd(chdir_path)

                    if chdir_path not in cache:
                        cache[chdir_path] = ftp.nlst()

                    cache_match = "ice_conc_{}_ease*_{:04d}{:02d}{:02d}*.nc".\
                        format(hs, el.year, el.month, el.day)
                    ftp_files = [el for el in cache[chdir_path]
                                 if fnmatch.fnmatch(el, cache_match)]

                    if len(ftp_files) > 1:
                        raise ValueError("More than a single file found: {}".
                                         format(ftp_files))
                    elif not len(ftp_files):
                        logging.warning("File is not available: {}".
                                        format(cache_match))
                        continue
                except ftplib.error_perm:
                    logging.warning("FTP error, possibly missing month chdir "
                                    "for {}".format(date_str))
                    continue

                with open(temp_path, "wb") as fh:
                    ftp.retrbinary("RETR {}".format(ftp_files[0]), fh.write)

                logging.debug("Downloaded {}".format(temp_path))
                data_files.append(temp_path)

        if ftp:
            ftp.quit()

        logging.debug("Files being processed: {}".format(data_files))

        if len(data_files):
            ds = xr.open_mfdataset(data_files,
                                   combine="nested",
                                   concat_dim="time",
                                   data_vars=["ice_conc"],
                                   drop_variables=var_remove_list,
                                   engine="netcdf4")

            logging.debug("Processing out extraneous data")

            ds = ds.drop_vars(var_remove_list, errors="ignore")
            da = ds.resample(time="1D").mean().ice_conc

            da = da.where(da < 9.9e+36, 0.)  # Missing values
            da /= 100.  # Convert from SIC % to fraction

            if 'lat' not in da.coords:
                raise RuntimeError("latitude missing, fix required that has "
                                   "been removed in this version")
                # TODO: ref another file if this is missing, but hopefully the
                #  coordinates will be projected from mfdataset
                # logging.warning("Adding lat vals to coords, as missing in "
                #                "this set: {}".format(file))
                # da.coords['lat'] = lat_vals

            # In experimenting, I don't think this is actually required
            for month, mask in self._mask_dict.items():
                da.loc[dict(time=(da['time.month'] == month))].values[:, ~mask] = 0.

            for date in da.time.values:
                day_da = da.sel(time=slice(date, date))

                if np.sum(np.isnan(day_da.data)) > 0:
                    logging.warning("NaNs detected, adding to invalid "
                                    "list: {}".format(date))
                    self._invalid_dates.append(pd.to_datetime(date))

            var_folder = self.get_data_var_folder(var)
            group_by = "time.year"

            # xr.concat([ds, ds2], dim="time").sortby("time")
            for year, year_da in da.groupby(group_by):
                req_date = pd.to_datetime(year_da.time.values[0])
                year_path = os.path.join(
                    var_folder, "{}.nc".format(
                        getattr(req_date, "year")))

                logging.info("Saving {}".format(year_path))
                year_da.compute()
                year_da.to_netcdf(year_path)

        self.missing_dates()

        if self._delete:
            for fpath in data_files:
                os.unlink(fpath)

    def missing_dates(self):
        """

        :return:
        """
        filenames = set([os.path.join(
            self.get_data_var_folder("siconca"),
            "{}.nc".format(el.strftime("%Y")))
            for el in self._dates])
        filenames = [f for f in filenames if os.path.exists(f)]

        logging.info("Opening for interpolation: {}".format(filenames))
        ds = xr.open_mfdataset(filenames,
                               combine="nested",
                               concat_dim="time",
                               parallel=True)
        return self._missing_dates(ds.ice_conc)

    def _missing_dates(self, da: object) -> object:
        """

        :param da:
        :return:
        """
        if pd.Timestamp(1979, 1, 2) in da.time.values \
                and dt.date(1979, 1, 1) in self._dates\
                and pd.Timestamp(1979, 1, 1) not in da.time.values:
            da_1979_01_01 = da.sel(
                time=[pd.Timestamp(1979, 1, 2)]).copy().assign_coords(
                {'time': [pd.Timestamp(1979, 1, 1)]})
            da = xr.concat([da, da_1979_01_01], dim='time')
            da = da.sortby('time')

        dates_obs = [pd.to_datetime(date).date() for date in da.time.values]
        dates_all = [pd.to_datetime(date).date() for date in
                     pd.date_range(min(self._dates), max(self._dates))]

        # Weirdly, we were getting future warnings for timestamps, but unsure
        # where from
        invalid_dates = [pd.to_datetime(d).date() for d in self._invalid_dates]
        missing_dates = [date for date in dates_all
                         if date not in dates_obs
                         or date in invalid_dates]

        logging.info("Processing {} missing dates".format(len(missing_dates)))

        missing_dates_path = os.path.join(
            self.get_data_var_folder("siconca"), "missing_days.csv")

        with open(missing_dates_path, "a") as fh:
            for date in missing_dates:
                # FIXME: slightly unusual format for Ymd dates
                fh.write(date.strftime("%Y,%m,%d\n"))

        logging.debug("Interpolating {} missing dates".
                      format(len(missing_dates)))

        for date in missing_dates:
            # TODO: test, but overcomes issue with reprocessing
            if pd.Timestamp(date) not in da.time.values:
                logging.info("Interpolating {}".format(date))
                da = xr.concat([da,
                                da.interp(time=pd.to_datetime(date))],
                               dim='time')

        logging.debug("Finished interpolation")

        da = da.sortby('time')
        da.data = np.array(da.data, dtype=self._dtype)

        for date in missing_dates:
            date_str = pd.to_datetime(date).strftime("%Y_%m_%d")
            fpath = os.path.join(
                self.get_data_var_folder(
                    "siconca", append=[str(pd.to_datetime(date).year)]),
                "missing.{}.nc".format(date_str))

            if not os.path.exists(fpath):
                day_da = da.sel(time=slice(date, date))
                mask = self._mask_dict[pd.to_datetime(date).month]

                day_da.data[0][~mask] = 0.

                logging.info("Writing missing date file {}".format(fpath))
                day_da.to_netcdf(fpath)

        return da


def main():
    args = download_args(var_specs=False)

    logging.info("OSASIF-SIC Data Downloading")
    sic = SICDownloader(
        dates=[pd.to_datetime(date).date() for date in
               pd.date_range(args.start_date, args.end_date, freq="D")],
        delete_tempfiles=args.delete,
        north=args.hemisphere == "north",
        south=args.hemisphere == "south",
    )
    sic.download()



import datetime as dt
import os

from icenet.data.cli import process_args, process_date_args
from icenet.data.process import IceNetPreProcessor
from icenet.data.sic.mask import Masks
from icenet.data.processors.utils import sic_interpolate

"""

"""


class IceNetOSIPreProcessor(IceNetPreProcessor):
    """

    :param missing_dates:
    """
    def __init__(self, *args,
                 missing_dates: object = None,
                 **kwargs):
        super().__init__(*args, identifier="osisaf", **kwargs)

        missing_dates_path = os.path.join(
            self._source_data,
            "siconca",
            "missing_days.csv")

        missing_dates = [] if missing_dates is None else missing_dates
        assert type(missing_dates) is list

        with open(missing_dates_path, "r") as fh:
            missing_dates += [dt.date(*[int(s)
                                        for s in line.strip().split(",")])
                              for line in fh.readlines()]
        self.missing_dates = list(set(missing_dates))

    def pre_normalisation(self,
                          var_name: str,
                          da: object):
        """

        :param var_name:
        :param da:
        :return:
        """
        if var_name != "siconca":
            raise RuntimeError("OSISAF SIC implementation should be dealing "
                               "with siconca, ")
        else:
            masks = Masks(north=self.north, south=self.south)
            return sic_interpolate(da, masks)


def main():
    args = process_args()
    dates = process_date_args(args)

    osi = IceNetOSIPreProcessor(
        args.abs,
        args.anom,
        args.name,
        dates["train"],
        dates["val"],
        dates["test"],
        linear_trends=args.trends,
        linear_trend_steps=args.trend_lead,
        north=args.hemisphere == "north",
        ref_procdir=args.ref,
        south=args.hemisphere == "south"
    )
    osi.init_source_data(
        lag_days=args.lag,
    )
    osi.process()



import tensorflow as tf


class WeightedMSE(tf.keras.losses.MeanSquaredError):
    """Custom keras loss/metric for mean squared error

    :param name:
    """

    def __init__(self,
                 name: str = 'mse', **kwargs):
        super().__init__(name=name, **kwargs)

    def __call__(self,
                 y_true: object,
                 y_pred: object,
                 sample_weight: object = None):
        """
        :param y_true: Ground truth outputs
        :param y_pred: Network predictions
        :param sample_weight: Pixelwise mask weighting for metric summation
        :return: Mean squared error of SIC (%) (float)
        """

        # TF automatically reduces along final dimension - include dummy axis
        y_true = tf.expand_dims(y_true, axis=-1)
        y_pred = tf.expand_dims(y_pred, axis=-1)

        #if sample_weight is not None:
        #    sample_weight = tf.expand_dims(sample_weight, axis=-1)

        return super().__call__(100*y_true, 100*y_pred, sample_weight=sample_weight)



import logging

from icenet.data.cli import process_args, process_date_args
from icenet.data.process import IceNetPreProcessor
from icenet.data.sic.mask import Masks
from icenet.data.processors.utils import sic_interpolate

"""

"""


class IceNetCMIPPreProcessor(IceNetPreProcessor):
    """

    :param source:
    :param member:
    """
    def __init__(self,
                 source: str,
                 member: str,
                 *args, **kwargs):
        cmip_source = "{}.{}".format(source, member)
        super().__init__(*args,
                         identifier="cmip6.{}".format(cmip_source),
                         **kwargs)

    def pre_normalisation(self,
                          var_name: str,
                          da: object):
        """

        :param var_name:
        :param da:
        :return:
        """
        if var_name == "siconca":
            masks = Masks(north=self.north, south=self.south)
            return sic_interpolate(da, masks)

        return da


def main():
    args = process_args(
        extra_args=[
            (["source"], dict(type=str)),
            (["member"], dict(type=str)),
        ],
    )
    dates = process_date_args(args)

    cmip = IceNetCMIPPreProcessor(
        args.source,
        args.member,
        args.abs,
        args.anom,
        args.name,
        dates["train"],
        dates["val"],
        dates["test"],
        linear_trends=args.trends,
        linear_trend_days=args.trend_lead,
        north=args.hemisphere == "north",
        ref_procdir=args.ref,
        south=args.hemisphere == "south",
        update_key=args.update_key,
    )
    cmip.init_source_data(
        lag_days=args.lag,
    )
    cmip.process()



import datetime as dt
import logging
import os
import time

import dask
import dask.array as da

from dask.distributed import Client, LocalCluster

import numpy as np
import pandas as pd
import tensorflow as tf
import xarray as xr

from icenet.data.process import IceNetPreProcessor
from icenet.data.loaders.base import IceNetBaseDataLoader
from icenet.data.loaders.utils import IceNetDataWarning, write_tfrecord
from icenet.data.sic.mask import Masks


"""
Dask implementations for icenet data loading

Still WIP to re-introduce alternate implementations that might work better in 
certain deployments

"""


class DaskBaseDataLoader(IceNetBaseDataLoader):
    def __init__(self,
                 *args,
                 dask_port: int = 8888,
                 dask_timeouts: int = 60,
                 dask_tmp_dir: object = "/tmp",
                 **kwargs):
        super().__init__(*args, **kwargs)

        self._dashboard_port = dask_port
        self._timeout = dask_timeouts
        self._tmp_dir = dask_tmp_dir

    def generate(self):
        """

        """
        dashboard = "localhost:{}".format(self._dashboard_port)

        with dask.config.set({
            "temporary_directory": self._tmp_dir,
            "distributed.comm.timeouts.connect": self._timeout,
            "distributed.comm.timeouts.tcp": self._timeout,
        }):
            cluster = LocalCluster(
                dashboard_address=dashboard,
                n_workers=self.workers,
                threads_per_worker=1,
                scheduler_port=0,
            )
            logging.info("Dashboard at {}".format(dashboard))

            with Client(cluster) as client:
                logging.info("Using dask client {}".format(client))
                self.client_generate(client,
                                     dates_override=self.dates_override,
                                     pickup=self.pickup)

    def client_generate(self,
                        client: object,
                        dates_override: object = None,
                        pickup: bool = False):
        """

        :param client:
        :param dates_override:
        :param pickup:
        """
        raise NotImplementedError("generate called on non-implementation")


class DaskMultiSharingWorkerLoader(DaskBaseDataLoader):
    def __init__(self,
                 *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        # TODO: https://github.com/icenet-ai/icenet/blob/83fdbf4b23ccf6ac221e77809b47d407b70b707f/icenet2/data/loader.py
        raise NotImplementedError("Not yet adapted from old implementation")

    def client_generate(self,
                        client: object,
                        dates_override: object = None,
                        pickup: bool = False):
        """

        :param client:
        :param dates_override:
        :param pickup:
        """
        pass

    def generate_sample(self,
                        date: object,
                        prediction: bool = False):
        """

        :param date:
        :param prediction:
        """
        pass


class DaskMultiWorkerLoader(DaskBaseDataLoader):
    def __init__(self,
                 *args,
                 futures_per_worker: int = 2,
                 **kwargs):
        super().__init__(*args, **kwargs)

        masks = Masks(north=self.north, south=self.south)
        self._masks = da.array([
            masks.get_active_cell_mask(month) for month in range(1, 13)])

        self._futures = futures_per_worker

    def client_generate(self,
                        client: object,
                        dates_override: object = None,
                        pickup: bool = False):
        """

        :param client:
        :param dates_override:
        :param pickup:
        """
        # TODO: for each set, validate every variable has an appropriate file
        #  in the configuration arrays, otherwise drop the forecast date
        splits = ("train", "val", "test")

        if dates_override and type(dates_override) is dict:
            for split in splits:
                assert split in dates_override.keys() \
                       and type(dates_override[split]) is list, \
                       "{} needs to be list in dates_override".format(split)
        elif dates_override:
            raise RuntimeError("dates_override needs to be a dict if supplied")

        counts = {el: 0 for el in splits}
        exec_times = []

        def batch(batch_dates, num):
            i = 0
            while i < len(batch_dates):
                yield batch_dates[i:i + num]
                i += num

        masks = client.scatter(self._masks, broadcast=True)

        for dataset in splits:
            batch_number = 0
            futures = []

            forecast_dates = set([dt.datetime.strptime(s,
                                  IceNetPreProcessor.DATE_FORMAT).date()
                                  for identity in
                                  self._config["sources"].keys()
                                  for s in
                                  self._config["sources"][identity]
                                  ["dates"][dataset]])

            if dates_override:
                logging.info("{} available {} dates".
                             format(len(forecast_dates), dataset))
                forecast_dates = forecast_dates.intersection(
                    dates_override[dataset])
            forecast_dates = sorted(list(forecast_dates))

            output_dir = self.get_data_var_folder(dataset)
            tf_path = os.path.join(output_dir, "{:08}.tfrecord")

            logging.info("{} {} dates to process, generating cache "
                         "data.".format(len(forecast_dates), dataset))

            for dates in batch(forecast_dates, self._output_batch_size):
                if not pickup or \
                    (pickup and
                     not os.path.exists(tf_path.format(batch_number))):
                    args = [
                        self._channels,
                        self._dtype,
                        self._loss_weight_days,
                        self._meta_channels,
                        self._missing_dates,
                        self._n_forecast_days,
                        self.num_channels,
                        self._shape,
                        self._trend_steps,
                        masks,
                        False
                    ]

                    fut = client.submit(generate_and_write,
                                        tf_path.format(batch_number),
                                        self.get_sample_files(),
                                        dates,
                                        args,
                                        dry=self._dry)
                    futures.append(fut)

                    # Use this to limit the future list, to avoid crashing the
                    # distributed scheduler / workers (task list gets too big!)
                    if len(futures) >= self._workers * self._futures:
                        for tf_data, samples, gen_times \
                                in client.gather(futures):
                            logging.info("Finished output {}".format(tf_data))
                            counts[dataset] += samples
                            exec_times += gen_times
                        futures = []

                    # tf_data, samples, times = generate_and_write(
                    #    tf_path.format(batch_number), args, dry=self._dry)
                else:
                    counts[dataset] += len(dates)
                    logging.warning("Skipping {} on pickup run".
                                    format(tf_path.format(batch_number)))

                batch_number += 1

            # Hoover up remaining futures
            for tf_data, samples, gen_times \
                    in client.gather(futures):
                logging.info("Finished output {}".format(tf_data))
                counts[dataset] += samples
                exec_times += gen_times

        if len(exec_times) > 0:
            logging.info("Average sample generation time: {}".
                         format(np.average(exec_times)))
        self._write_dataset_config(counts)

    def generate_sample(self,
                        date: object,
                        prediction: bool = False):
        """

        :param date:
        :param prediction:
        :return:
        """

        ds_kwargs = dict(
            chunks=dict(time=1, yc=self._shape[0], xc=self._shape[1]),
            drop_variables=["month", "plev", "realization"],
            parallel=True,
        )
        var_files = self.get_sample_files()

        var_ds = xr.open_mfdataset(
            [v for k, v in var_files.items()
             if k not in self._meta_channels
             and not k.endswith("linear_trend")],
            **ds_kwargs)
        var_ds = var_ds.transpose("yc", "xc", "time")

        trend_files = \
            [v for k, v in var_files.items()
             if k.endswith("linear_trend")]
        trend_ds = None
        
        if len(trend_files) > 0:
            trend_ds = xr.open_mfdataset(
                trend_files,
                **ds_kwargs)

            trend_ds = trend_ds.transpose("yc", "xc", "time")

        args = [
            self._channels,
            self._dtype,
            self._loss_weight_days,
            self._meta_channels,
            self._missing_dates,
            self._n_forecast_days,
            self.num_channels,
            self._shape,
            self._trend_steps,
            self._masks,
            prediction
        ]

        x, y, sw = generate_sample(date,
                                   var_ds,
                                   var_files,
                                   trend_ds,
                                   *args)
        return x.compute(), y.compute(), sw.compute()


def generate_and_write(path: str,
                       var_files: object,
                       dates: object,
                       args: tuple,
                       dry: bool = False):
    """

    :param path:
    :param var_files:
    :param dates:
    :param args:
    :param dry:
    :return:
    """
    count = 0
    times = []

    # TODO: refactor, this is very smelly - with new data throughput args
    #  will always be the same
    (channels,
     dtype,
     loss_weight_days,
     meta_channels,
     missing_dates,
     n_forecast_days,
     num_channels,
     shape,
     trend_steps,
     masks,
     prediction) = args

    ds_kwargs = dict(
        chunks=dict(time=1, yc=shape[0], xc=shape[1]),
        drop_variables=["month", "plev", "realization"],
        parallel=True,
    )

    var_ds = xr.open_mfdataset(
        [v for k, v in var_files.items()
         if k not in meta_channels and not k.endswith("linear_trend")],
        **ds_kwargs)
    var_ds = var_ds.transpose("yc", "xc", "time")

    trend_files = [v for k, v in var_files.items()
                   if k.endswith("linear_trend")]
    trend_ds = None

    if len(trend_files):
        trend_ds = xr.open_mfdataset(
            trend_files,
            **ds_kwargs)
        trend_ds = trend_ds.transpose("yc", "xc", "time")

    with tf.io.TFRecordWriter(path) as writer:
        for date in dates:
            start = time.time()

            try:
                x, y, sample_weights = generate_sample(date,
                                                       var_ds,
                                                       var_files,
                                                       trend_ds,
                                                       *args)
                if not dry:
                    x[da.isnan(x)] = 0.

                    x, y, sample_weights = dask.compute(x, y, sample_weights,
                                                        optimize_graph=True)
                    write_tfrecord(writer,
                                   x, y, sample_weights)
                count += 1
            except IceNetDataWarning:
                continue

            end = time.time()
            times.append(end - start)
            logging.debug("Time taken to produce {}: {}".
                          format(date, times[-1]))
    return path, count, times


def generate_sample(forecast_date: object,
                    var_ds: object,
                    var_files: object,
                    trend_ds: object,
                    channels: object,
                    dtype: object,
                    loss_weight_days: bool,
                    meta_channels: object,
                    missing_dates: object,
                    n_forecast_days: int,
                    num_channels: int,
                    shape: object,
                    trend_steps: object,
                    masks: object,
                    prediction: bool = False):
    """


    :param forecast_date:
    :param var_ds:
    :param var_files:
    :param trend_ds:
    :param channels:
    :param dtype:
    :param loss_weight_days:
    :param meta_channels:
    :param missing_dates:
    :param n_forecast_days:
    :param num_channels:
    :param shape:
    :param trend_steps:
    :param masks:
    :param prediction:
    :return:
    """

    # Prepare data sample
    # To become array of shape (*raw_data_shape, n_forecast_days)
    forecast_dts = [forecast_date + dt.timedelta(days=n)
                    for n in range(n_forecast_days)]

    y = da.zeros((*shape, n_forecast_days, 1), dtype=dtype)
    sample_weights = da.zeros((*shape, n_forecast_days, 1), dtype=dtype)

    if not prediction:
        try:
            sample_output = var_ds.siconca_abs.sel(time=forecast_dts)
        except KeyError as sic_ex:
            logging.exception("Issue selecting data for non-prediction sample, "
                              "please review siconca ground-truth: dates {}".
                              format(forecast_dts))
            raise RuntimeError(sic_ex)
        y[:, :, :, 0] = sample_output

    # Masked recomposition of output
    for leadtime_idx in range(n_forecast_days):
        forecast_day = forecast_date + dt.timedelta(days=leadtime_idx)

        if any([forecast_day == missing_date
                for missing_date in missing_dates]):
            sample_weight = da.zeros(shape, dtype)
        else:
            # Zero loss outside of 'active grid cells'
            sample_weight = masks[forecast_day.month - 1]
            sample_weight = sample_weight.astype(dtype)

            # We can pick up nans, which messes up training
            sample_weight[da.isnan(y[..., leadtime_idx, 0])] = 0

            # Scale the loss for each month s.t. March is
            #   scaled by 1 and Sept is scaled by 1.77
            if loss_weight_days:
                sample_weight *= 33928. / sample_weight.sum()

        sample_weights[:, :, leadtime_idx, 0] = sample_weight

    # INPUT FEATURES
    x = da.zeros((*shape, num_channels), dtype=dtype)
    v1, v2 = 0, 0

    for var_name, num_channels in channels.items():
        if var_name in meta_channels:
            continue

        v2 += num_channels

        if var_name.endswith("linear_trend"):
            channel_ds = trend_ds
            if type(trend_steps) == list:
                channel_dates = [pd.Timestamp(forecast_date +
                                              dt.timedelta(days=int(n)))
                                 for n in trend_steps]
            else:
                channel_dates = [pd.Timestamp(forecast_date +
                                              dt.timedelta(days=n))
                                 for n in range(num_channels)]
        else:
            channel_ds = var_ds
            channel_dates = [pd.Timestamp(forecast_date - dt.timedelta(days=n))
                             for n in range(num_channels)]

        channel_data = []
        for cdate in channel_dates:
            try:
                channel_data.append(getattr(channel_ds, var_name).
                                    sel(time=cdate))
            except KeyError:
                channel_data.append(da.zeros(shape))

        x[:, :, v1:v2] = da.from_array(channel_data).transpose([1, 2, 0])
        v1 += num_channels

    for var_name in meta_channels:
        if channels[var_name] > 1:
            raise RuntimeError("{} meta variable cannot have more than "
                               "one channel".format(var_name))

        meta_ds = xr.open_dataarray(var_files[var_name])

        if var_name in ["sin", "cos"]:
            ref_date = "2012-{}-{}".format(forecast_date.month,
                                           forecast_date.day)
            trig_val = meta_ds.sel(time=ref_date).to_numpy()
            x[:, :, v1] = da.broadcast_to([trig_val], shape)
        else:
            x[:, :, v1] = da.array(meta_ds.to_numpy())
        v1 += channels[var_name]

    #    x.visualize(filename='x.svg', optimize_graph=True)
    #    y.visualize(filename='y.svg', optimize_graph=True)
    #    sample_weights.visualize(filename='sample_weights.svg', optimize_graph=True)
    #    import sys
    #    sys.exit(0)

    return x, y, sample_weights






import datetime as dt
import logging
import os
import shutil

import numpy as np
import pandas as pd
import xarray as xr

from icenet.data.cli import download_args
from icenet.data.producers import Generator
from icenet.utils import run_command
from icenet.data.sic.utils import SIC_HEMI_STR

"""Sea Ice Masks

"""


class Masks(Generator):
    """

    :param polarhole_dates:
    :param polarhole_radii:
    :param data_shape:
    :param dtype:
    """

    LAND_MASK_FILENAME = "land_mask.npy"
    # FIXME: nh/sh?
    POLARHOLE_RADII = (28, 11, 3)
    POLARHOLE_DATES = (
        dt.date(1987, 6, 1),
        dt.date(2005, 10, 1),
        dt.date(2015, 12, 1),
    )

    def __init__(self, *args,
                 polarhole_dates: object = POLARHOLE_DATES,
                 polarhole_radii: object = POLARHOLE_RADII,
                 data_shape: object = (432, 432),
                 dtype: object = np.float32,
                 **kwargs):
        super().__init__(*args, identifier="masks", **kwargs)

        self._polarhole_dates = polarhole_dates
        self._polarhole_radii = polarhole_radii
        self._dtype = dtype
        self._shape = data_shape

        self.init_params()

    def init_params(self):
        """

        """
        params_path = os.path.join(
            self.get_data_var_folder("masks"),
            "masks.params"
        )

        if not os.path.exists(params_path):
            with open(params_path, "w") as fh:
                for i, polarhole in enumerate(self._polarhole_radii):
                    fh.write("{}\n".format(
                        ",".join([str(polarhole),
                                  self._polarhole_dates[i].strftime("%Y%m%d")]
                                 )))
        else:
            lines = [el.strip().split(",")
                     for el in open(params_path, "r").readlines()]
            radii, dates = zip(*lines)
            self._polarhole_dates = [dt.datetime.strptime(el, "%Y%m%d").date()
                                     for el in dates]
            self._polarhole_radii = [int(r) for r in radii]

    def generate(self,
                 year: int = 2000,
                 save_land_mask: bool = True,
                 save_polarhole_masks: bool = True,
                 remove_temp_files: bool = False):
        """Generate a set of data masks

        :param year:
        :param save_land_mask:
        :param save_polarhole_masks:
        :param remove_temp_files:
        """
        siconca_folder = self.get_data_var_folder("siconca")

        # FIXME: cut-dirs can be change intolerant, better use -O, changed
        #  from 4 to 5
        retrieve_cmd_template_osi450 = \
            "wget -m -nH --cut-dirs=4 -P {} " \
            "ftp://osisaf.met.no/reprocessed/ice/conc/v2p0/{:04d}/{:02d}/{}"
        filename_template_osi450 = \
            'ice_conc_{}_ease2-250_cdr-v2p0_{:04d}{:02d}021200.nc'

        # Generate the land-lake-sea mask using the second day from each month
        # of the year 2000 (chosen arbitrarily as the mask is fixed within
        # month)
        for month in range(1, 13):
            # Download the data if not already downloaded
            filename_osi450 = filename_template_osi450.format(
                SIC_HEMI_STR[self.hemisphere_str[0]], year, month)

            month_str = '{:02d}'.format(month)
            month_folder = os.path.join(siconca_folder, str(year), month_str)
            month_path = os.path.join(month_folder, filename_osi450)

            if not os.path.exists(month_path):
                run_command(retrieve_cmd_template_osi450.format(
                    siconca_folder, year, month, filename_osi450))
            else:
                logging.info("siconca {} already exists".
                             format(filename_osi450))

            with xr.open_dataset(month_path) as ds:
                status_flag = ds['status_flag']
                status_flag = np.array(status_flag.data).astype(np.uint8)
                status_flag = status_flag.reshape(*self._shape)

                binary = np.unpackbits(status_flag, axis=1).\
                    reshape(*self._shape, 8)

                # Mask out: land, lake, and 'outside max climatology' (open sea)
                max_extent_mask = np.sum(
                    binary[:, :, [7, 6, 0]], axis=2).reshape(*self._shape) >= 1
                max_extent_mask = ~max_extent_mask

                # FIXME: Remove Caspian and Black seas - should we do this sh?
                if self.north:
                    max_extent_mask[325:386, 317:380] = False

            mask_path = os.path.join(self.get_data_var_folder("masks"),
                                     "active_grid_cell_mask_{:02d}.npy".
                                     format(month))
            logging.info("Saving {}".format(mask_path))

            np.save(mask_path, max_extent_mask)

            land_mask_path = os.path.join(self.get_data_var_folder("masks"),
                                          Masks.LAND_MASK_FILENAME)

            if save_land_mask and \
                    not os.path.exists(land_mask_path) \
                    and month == 1:
                land_mask = np.sum(binary[:, :, [7, 6]], axis=2).\
                                reshape(*self._shape) >= 1

                logging.info("Saving {}".format(land_mask_path))
                np.save(land_mask_path, land_mask)

        # Delete the data/siconca/2000 folder holding the temporary daily files
        if remove_temp_files:
            logging.info("Removing {}".format(siconca_folder))
            shutil.rmtree(siconca_folder)

        if save_polarhole_masks and not self.south:
            # Generate the polar hole masks
            x = np.tile(np.arange(0, self._shape[1]).
                        reshape(self._shape[0], 1), (1, self._shape[1])).\
                astype(self._dtype) - 215.5
            y = np.tile(np.arange(0, self._shape[1]).
                        reshape(1, self._shape[1]), (self._shape[0], 1)).\
                astype(self._dtype) - 215.5
            squaresum = np.square(x) + np.square(y)

            for i, radius in enumerate(self._polarhole_radii):
                polarhole = np.full(self._shape, False)
                polarhole[squaresum < radius**2] = True

                polarhole_path = os.path.join(self.get_data_var_folder("masks"),
                                              "polarhole{}_mask.npy".
                                              format(i+1))
                logging.info("Saving polarhole {}".format(polarhole_path))
                np.save(polarhole_path, polarhole)

    def get_active_cell_mask(self,
                             month: object) -> object:
        """

        :param month:
        :return:
        """
        mask_path = os.path.join(self.get_data_var_folder("masks"),
                                 "active_grid_cell_mask_{:02d}.npy".
                                 format(month))

        if not os.path.exists(mask_path):
            raise RuntimeError("Active cell masks have not been generated, "
                               "this is not done automatically so you might "
                               "want to address this!")

        # logging.debug("Loading active cell mask {}".format(mask_path))
        return np.load(mask_path)

    def get_active_cell_da(self,
                           src_da: object) -> object:
        """

        :param src_da:
        """

        return xr.DataArray(
            [self.get_active_cell_mask(pd.to_datetime(date).month)
             for date in src_da.time.values],
            dims=('time', 'yc', 'xc'),
            coords={
                'time': src_da.time.values,
                'yc': src_da.yc.values,
                'xc': src_da.xc.values,
            }
        )

    def get_land_mask(self,
                      land_mask_filename: str = LAND_MASK_FILENAME) -> object:
        """

        :param land_mask_filename:
        :return:
        """
        mask_path = os.path.join(self.get_data_var_folder("masks"),
                                 land_mask_filename)

        if not os.path.exists(mask_path):
            raise RuntimeError("Land mask has not been generated, this is "
                               "not done automatically so you might want to "
                               "address this!")

        # logging.debug("Loading land mask {}".format(mask_path))
        return np.load(mask_path)

    def get_polarhole_mask(self,
                           date: object) -> object:
        """

        :param date:
        :return:
        """
        if self.south:
            return None

        for i, r in enumerate(self._polarhole_radii):
            if date <= self._polarhole_dates[i]:
                polarhole_path = os.path.join(self.get_data_var_folder("masks"),
                                              "polarhole{}_mask.npy".
                                              format(i + 1))
                # logging.debug("Loading polarhole {}".format(polarhole_path))
                return np.load(polarhole_path)
        return None

    def get_blank_mask(self) -> object:
        """

        :return:
        """
        return np.full(self._shape, False)


def main():
    args = download_args(dates=False, var_specs=False)

    north = args.hemisphere == "north"
    south = args.hemisphere == "south"

    Masks(north=north, south=south).\
        generate(save_polarhole_masks=north)



from abc import abstractmethod, ABCMeta
from pprint import pformat

import collections
import datetime as dt
import glob
import logging
import os
import re

import pandas as pd

from icenet.utils import Hemisphere, HemisphereMixin


class DataCollection(HemisphereMixin, metaclass=ABCMeta):
    """

    :param identifier:
    :param north:
    :param south:
    :param path:
    """

    @abstractmethod
    def __init__(self, *args,
                 identifier: object = None,
                 north: bool = True,
                 south: bool = False,
                 path: str = os.path.join(".", "data"),
                 **kwargs):
        self._identifier = identifier
        self._path = os.path.join(path, identifier)
        self._hemisphere = (Hemisphere.NORTH if north else Hemisphere.NONE) | \
                           (Hemisphere.SOUTH if south else Hemisphere.NONE)

        assert self._identifier, "No identifier supplied"
        assert self._hemisphere != Hemisphere.NONE, "No hemispheres selected"

    @property
    def base_path(self):
        return self._path

    @base_path.setter
    def base_path(self, path):
        self._path = path

    @property
    def identifier(self):
        return self._identifier


class DataProducer(DataCollection):
    """
    :param dry:
    :param overwrite:
    """
    def __init__(self, *args,
                 dry: bool = False,
                 overwrite: bool = False,
                 **kwargs):
        super(DataProducer, self).__init__(*args, **kwargs)

        self.dry = dry
        self.overwrite = overwrite

        if os.path.exists(self._path):
            logging.debug("{} already exists".format(self._path))
        else:
            if not os.path.islink(self._path):
                logging.info("Creating path: {}".format(self._path))
                os.makedirs(self._path, exist_ok=True)
            else:
                logging.info("Skipping creation for symlink: {}".format(
                    self._path))

        # NOTE: specific limitation for the DataProducers, they'll only do one
        # hemisphere per instance
        assert self._hemisphere != Hemisphere.BOTH, "Both hemispheres selected"

    def get_data_var_folder(self,
                            var: str,
                            append: object = None,
                            hemisphere: object = None,
                            missing_error: bool = False) -> str:
        """

        :param var:
        :param append:
        :param hemisphere:
        :param missing_error:
        :return:
        """
        if not append:
            append = []

        if not hemisphere:
            # We can make the assumption because this implementation is limited
            # to a single hemisphere
            hemisphere = self.hemisphere_str[0]

        data_var_path = os.path.join(
            self.base_path, *[hemisphere, var, *append]
        )

        if not os.path.exists(data_var_path):
            if not missing_error:
                os.makedirs(data_var_path, exist_ok=True)
            else:
                raise OSError("Directory {} is missing and this is "
                              "flagged as an error!".format(data_var_path))

        return data_var_path


class Downloader(DataProducer):
    """

    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @abstractmethod
    def download(self):
        """

        """
        raise NotImplementedError("{}.download is abstract".
                                  format(__class__.__name__))


class Generator(DataProducer):
    """
    
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @abstractmethod
    def generate(self):
        """

        """
        raise NotImplementedError("{}.generate is abstract".
                                  format(__class__.__name__))


class Processor(DataProducer):
    """
    
    :param identifier: 
    :param source_data: 
    :param *args: 
    :param file_filters: 
    :param test_dates: 
    :param train_dates: 
    :param val_dates:
    """
    def __init__(self,
                 identifier: str,
                 source_data: object,
                 *args,
                 file_filters: object = (),
                 lead_time: int = 93,
                 test_dates: object = (),
                 train_dates: object = (),
                 val_dates: object = (),
                 **kwargs):
        super().__init__(*args,
                         identifier=identifier,
                         **kwargs)

        self._file_filters = list(file_filters)
        self._lead_time = lead_time
        self._source_data = os.path.join(source_data,
                                         identifier,
                                         self.hemisphere_str[0])
        self._var_files = dict()
        self._processed_files = dict()

        # TODO: better as a mixin?
        Dates = collections.namedtuple("Dates", ["train", "val", "test"])
        self._dates = Dates(train=list(train_dates),
                            val=list(val_dates),
                            test=list(test_dates))

    def init_source_data(self,
                         lag_days: object = None):
        """

        :param lag_days:
        """

        if not os.path.exists(self.source_data):
            raise OSError("Source data directory {} does not exist".
                          format(self.source_data))

        var_files = {}

        for date_category in ["train", "val", "test"]:
            dates = sorted(getattr(self._dates, date_category))

            if dates:
                logging.info("Processing {} dates for {} category".
                             format(len(dates), date_category))
            else:
                logging.info("No {} dates for this processor".
                             format(date_category))
                continue

            # TODO: ProcessPool for this (avoid the GIL for globbing)
            # FIXME: needs to deal with a lack of continuity in the date ranges
            if lag_days:
                logging.info("Including lag of {} days".format(lag_days))

                additional_lag_dates = []

                for date in dates:
                    for day in range(lag_days):
                        lag_date = date - dt.timedelta(days=day + 1)
                        if lag_date not in dates:
                            additional_lag_dates.append(lag_date)
                dates += list(set(additional_lag_dates))

            # FIXME: this is conveniently supplied for siconca_abs on
            #  training with OSISAF data, but are we exploiting the
            #  convenient usage of this data for linear trends?
            if self._lead_time:
                logging.info("Including lead of {} days".format(self._lead_time))

                additional_lead_dates = []

                for date in dates:
                    for day in range(self._lead_time):
                        lead_day = date + dt.timedelta(days=day + 1)
                        if lead_day not in dates:
                            additional_lead_dates.append(lead_day)
                dates += list(set(additional_lead_dates))

            globstr = "{}/**/[12]*.nc".format(self.source_data)

            logging.debug("Globbing {} from {}".format(date_category, globstr))
            dfs = glob.glob(globstr, recursive=True)
            logging.debug("Globbed {} files".format(len(dfs)))

            # FIXME: using hyphens broadly no?
            data_dates = [df.split(os.sep)[-1][:-3].replace("_", "-")
                          for df in dfs]
            dt_series = pd.Series(dfs, index=data_dates)

            logging.debug("Create structure of {} files".format(len(dt_series)))

            # Ensure we're ordered, it has repercussions for xarray
            for date in sorted(dates):
                try:
                    match_dfs = dt_series[date.strftime("%Y")]

                    if type(match_dfs) == str:
                        match_dfs = [match_dfs]
                except KeyError:
                    logging.info("No data found for {}, outside data boundary "
                                 "perhaps?".format(date.strftime("%Y-%m-%d")))
                    match_dfs = []

                for df in match_dfs:
                    if any([flt in os.path.split(df)[1]
                            for flt in self._file_filters]):
                        continue

                    path_comps = str(os.path.split(df)[0]).split(os.sep)
                    var = path_comps[-1]

                    # The year is in the path, fall back one further
                    if re.match(r'^\d{4}$', var):
                        var = path_comps[-2]

                    if var not in var_files.keys():
                        var_files[var] = list()

                    if df not in var_files[var]:
                        var_files[var].append(df)

        # TODO: allow option to ditch dates from train/val/test for missing
        #  var files
        self._var_files = {
            var: var_files[var] for var in sorted(var_files.keys())
        }
        for var in self._var_files.keys():
            logging.info("Got {} files for {}".format(
                len(self._var_files[var]), var))

    @abstractmethod
    def process(self):
        """

        """
        raise NotImplementedError("{}.process is abstract".
                                  format(__class__.__name__))

    def save_processed_file(self,
                            var_name: str,
                            name: str,
                            data: object, **kwargs):
        """

        :param var_name:
        :param name:
        :param data:
        :param kwargs:
        :return:
        """
        file_path = os.path.join(
            self.get_data_var_folder(var_name, **kwargs), name)
        data.to_netcdf(file_path)

        if var_name not in self._processed_files.keys():
            self._processed_files[var_name] = list()

        if file_path not in self._processed_files[var_name]:
            logging.debug("Adding {} file: {}".format(var_name, file_path))
            self._processed_files[var_name].append(file_path)
        else:
            logging.warning("{} already exists in {} processed list".
                            format(file_path, var_name))
        return file_path

    @property
    def dates(self):
        return self._dates

    @property
    def lead_time(self):
        return self._lead_time

    @property
    def processed_files(self):
        return self._processed_files

    @property
    def source_data(self):
        return self._source_data



import configparser
import logging
import os
import time

import numpy as np
import pandas as pd
import xarray as xr

from icenet.data.cli import download_args
from icenet.data.interfaces.downloader import ClimateDownloader
from icenet.utils import run_command

"""
DATASET: global-reanalysis-phy-001-031-grepv2-daily
FTP ENDPOINT: ftp://my.cmems-du.eu/Core/GLOBAL_REANALYSIS_PHY_001_031/global-reanalysis-phy-001-031-grepv2-daily/1993/01/

"""


class ORAS5Downloader(ClimateDownloader):
    """Climate downloader to provide ORAS5 reanalysis data from CMEMS API

    These aren't available for CMIP training at daily frequencies

    :param identifier: how to identify this dataset
    :param var_map: override the default ERA5Downloader.CDI_MAP variable map
    """
    ENDPOINTS = {
        # TODO: See #49 - not yet used
        "cas":  "https://cmems-cas.cls.fr/cas/login",
        "dap":  "https://my.cmems-du.eu/thredds/dodsC/{dataset}",
        "motu": "https://my.cmems-du.eu/motu-web/Motu",
    }

    VAR_MAP = {
        "thetao": "thetao_oras",    # sea_water_potential_temperature
        "so": "so_oras",            # sea_water_salinity
        "uo": "uo_oras",            # eastward_sea_water_velocity
        "vo": "vo_oras",            # northward_sea_water_velocity
        "zos": "zos_oras",          # sea_surface_height_above_geoid
        "mlotst": "mlotst_oras",    # ocean_mixed_layer_thickness_defined_by_sigma_theta
    }

    def __init__(self,
                 *args,
                 cred_file: str = os.path.expandvars("$HOME/.cmems.creds"),
                 dataset: str = "global-reanalysis-phy-001-031-grepv2-daily",
                 identifier: str = "oras5",
                 max_failures: int = 3,
                 service: str = "GLOBAL_REANALYSIS_PHY_001_031-TDS",
                 var_map: object = None,
                 **kwargs):
        super().__init__(*args,
                         identifier=identifier,
                         **kwargs)

        cp = configparser.ConfigParser(default_section="auth")
        cp.read(cred_file)
        self._creds = dict(cp["auth"])
        self._dataset = dataset
        self._max_failures = max_failures
        self._service = service
        self._var_map = var_map if var_map else ORAS5Downloader.VAR_MAP

        assert self._max_threads <= 8, "Too many request threads for ORAS5 " \
                                       "(max.8)"

        for var_name in self._var_names:
            assert var_name in self._var_map.keys(), \
                "{} not in ORAS5 var map".format(var_name)

        self.download_method = self._single_motu_download

    def postprocess(self,
                    var: str,
                    download_path: object):
        """

        :param var:
        :param download_path:
        """
        logging.info("Postprocessing {} to {}".format(var, download_path))
        ds = xr.open_dataset(download_path)

        da = getattr(ds, self._var_map[var]).rename(var)
        da = da.mean("depth").compute()
        da.to_netcdf(download_path)

    def _single_motu_download(self,
                              var: str,
                              level: object,
                              req_dates: int,
                              download_path: object):
        """Implements a single download from ... server
        :param var:
        :param level:
        :param req_dates:
        :param download_path:
        :return:

        """
        attempts = 1
        success = False

        cmd = \
            """motuclient --quiet --motu {} \
                    --service-id {} \
                    --product-id {} \
                    --longitude-min -180 \
                    --longitude-max 179.75 \
                    --latitude-min {} \
                    --latitude-max {} \
                    --date-min "{} 00:00:00" \
                    --date-max "{} 00:00:00" \
                    --depth-min 0.5056 \
                    --depth-max 0.5059 \
                    --variable {} \
                    --out-dir {} \
                    --out-name {} \
                    --user {} \
                    --pwd '{}' \
            """.format(self.ENDPOINTS['motu'],
                       self._service,
                       self._dataset,
                       self.hemisphere_loc[2],
                       self.hemisphere_loc[0],
                       req_dates[0].strftime("%Y-%m-%d"),
                       req_dates[-1].strftime("%Y-%m-%d"),
                       self._var_map[var],
                       os.path.split(download_path)[0],
                       os.path.split(download_path)[1],
                       self._creds['username'],
                       self._creds['password'])

        tic = time.time()
        while not success:
            logging.debug("Attempt {}".format(attempts))

            ret = run_command(cmd)
            if ret.returncode != 0 or not os.path.exists(download_path):
                attempts += 1
                if attempts > self._max_failures:
                    logging.error("Couldn't download {} between {} and {}".
                                  format(var, req_dates[0], req_dates[-1]))
                    break
                time.sleep(30)
            else:
                success = True

        if success:
            dur = time.time() - tic
            logging.debug("Done in {}m:{:.0f}s. ".format(np.floor(dur / 60),
                                                         dur % 60))
        return success

    def additional_regrid_processing(self,
                                     datafile: object,
                                     cube_ease: object) -> object:
        """

        :param datafile:
        :param cube_ease:
        :return:
        """
        cube_ease.data = np.ma.filled(cube_ease.data, fill_value=0.)
        return cube_ease


def main():
    args = download_args(workers=True, extra_args=(
        (("-n", "--do-not-download"),
         dict(dest="download", action="store_false", default=True)),
        (("-p", "--do-not-postprocess"),
         dict(dest="postprocess", action="store_false", default=True))))

    logging.info("ORAS5 Data Downloading")
    oras5 = ORAS5Downloader(
        var_names=args.vars,
        # TODO: currently hardcoded
        pressure_levels=[None for _ in args.vars],
        dates=[pd.to_datetime(date).date() for date in
               pd.date_range(args.start_date, args.end_date, freq="D")],
        delete_tempfiles=args.delete,
        download=args.delete,
        max_threads=args.workers,
        postprocess=args.postprocess,
        north=args.hemisphere == "north",
        south=args.hemisphere == "south",
    )
    oras5.download()
    oras5.regrid()




import os

import numpy as np
import pandas as pd
import xarray as xr

from icenet.data.cli import process_args
from icenet.data.process import IceNetPreProcessor
from icenet.data.sic.mask import Masks

"""

"""


class IceNetMetaPreProcessor(IceNetPreProcessor):
    """

    :param name:
    :param include_circday:
    :param include_land:
    """

    def __init__(self,
                 name: str,
                 include_circday: bool = True,
                 include_land: bool = True,
                 **kwargs):
        super().__init__(abs_vars=[],
                         anom_vars=[],
                         identifier="meta",
                         linear_trends=tuple(),
                         name=name,
                         test_dates=[],
                         train_dates=[],
                         val_dates=[],
                         **kwargs)

        self._include_circday = include_circday
        self._include_land = include_land

    def init_source_data(self,
                         lag_days: object = None,
                         lead_days: object = None):
        """

        :param lag_days:
        :param lead_days:
        """
        raise NotImplementedError("No need to execute implementation for meta")

    def process(self):
        """

        """
        if self._include_circday:
            self._save_circday()

        if self._include_land:
            self._save_land()

        # FIXME: this is a bit messy, clarify meta_vars and interface between
        #  processing and dataset generation
        if self._update_loader:
            self.update_loader_config()

    def _save_land(self):
        """

        """
        land_mask = Masks(north=self.north, south=self.south).get_land_mask()
        land_map = np.ones(self._data_shape, dtype=self._dtype)
        land_map[~land_mask] = -1.

        if "land" not in self._meta_vars:
            self._meta_vars.append("land")

        da = xr.DataArray(
            data=land_map,
            dims=["yc", "xc"],
            attrs=dict(description="IceNet land mask metadata")
        )
        land_path = self.save_processed_file("land", "land.nc", da)
        return land_path

    def _save_circday(self):
        """

        """
        cos = []
        sin = []
        paths = []

        for date in pd.date_range(start='2012-1-1', end='2012-12-31'):
            if self.north:
                circday = date.dayofyear
            else:
                circday = date.dayofyear + 365.25 / 2

            cos.append(np.cos(2 * np.pi * circday / 366, dtype=self._dtype))
            sin.append(np.sin(2 * np.pi * circday / 366, dtype=self._dtype))

        for var_name in ["sin", "cos"]:
            if var_name not in self._meta_vars:
                self._meta_vars.append(var_name)

            da = xr.DataArray(
                data=eval(var_name),
                dims=["time"],
                coords=dict(
                    time=pd.date_range(start='2012-1-1', end='2012-12-31')
                ),
                attrs=dict(
                    description="IceNet {} mask metadata".format(var_name))
            )
            paths.append(
                self.save_processed_file(var_name, "{}.nc".format(var_name),
                                         da))
        return paths


def main():
    args = process_args(dates=False, ref_option=False)

    IceNetMetaPreProcessor(
        args.name,
        north=args.hemisphere == "north",
        south=args.hemisphere == "south"
    ).process()



import argparse
import datetime as dt
import logging
import os
import re

from concurrent.futures import as_completed, ProcessPoolExecutor

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import xarray as xr

from matplotlib.animation import FuncAnimation
from mpl_toolkits.axes_grid1 import make_axes_locatable

from icenet.process.predict import get_refcube
from icenet.utils import setup_logging


# TODO: This can be a plotting or analysis util function elsewhere
def get_dataarray_from_files(files: object,
                             numpy: bool = False) -> object:
    """

    :param files:
    :param numpy:
    :return:
    """
    if not numpy:
        ds = xr.open_mfdataset(files)
        # TODO: We're relying on single variable files from downloaders
        #  so maybe allow a specifier for this for multi var files?
        da = ds.to_array(dim=list(ds.data_vars)[0])[0]
    else:
        first_file = np.load(files[0])
        arr = np.zeros((len(files), *first_file.shape))
        dates = []

        assert len(first_file.shape) == 2, \
            "Wrong number of dims for use in videos".\
                format(len(first_file.shape))

        for np_idx in range(0, len(files)):
            arr[np_idx] = np.load(files[np_idx])
            nom = os.path.basename(files[np_idx])

            # TODO: error handling
            date_match = re.search(r"(\d{4})_(\d{1,2})_(\d{1,2})", nom)
            dates.append(pd.to_datetime(
                dt.date(*[int(s) for s in date_match.groups()])))

        # FIXME: naive implementations abound
        path_comps = os.path.dirname(files[0]).split(os.sep)
        ref_cube = get_refcube("north" in path_comps, "south" in path_comps)
        var_name = path_comps[-2]

        da = xr.DataArray(
            data=arr,
            dims=("time", "yc", "xc"),
            coords=dict(
                time=[pd.Timestamp(d) for d in dates],
                xc=ref_cube.coord("projection_x_coordinate").points,
                yc=ref_cube.coord("projection_y_coordinate").points,
            ),
            name=var_name,
        )

    return da


def xarray_to_video(da: object,
                    fps: int,
                    video_path: object = None,
                    mask: object = None,
                    mask_type: str = 'contour',
                    clim: object = None,
                    crop: object = None,
                    data_type: str = 'abs',
                    video_dates: object = None,
                    cmap: object = 'viridis',
                    figsize: int = 12,
                    dpi: int = 150) -> object:

    """
    Generate video of an xarray.DataArray. Optionally input a list of
    `video_dates` to show, otherwise the full set of time coordiantes
    of the dataset is used.

    :param da: Dataset to create video of.
    :param video_path: Path to save the video to.
    :param fps: Frames per second of the video.
    :param mask: Boolean mask with True over masked elements to overlay
    as a contour or filled contour. Defaults to None (no mask plotting).
    :param mask_type: 'contour' or 'contourf' dictating whether the mask is
    overlaid as a contour line or a filled contour.
    :param data_type: 'abs' or 'anom' describing whether the data is in absolute
    or anomaly format. If anomaly, the colorbar is centred on 0.
    :param video_dates: List of Pandas Timestamps or datetime.datetime objects
    to plot video from the dataset.
    :param crop: [(a, b), (c, d)] to crop the video from a:b and c:d
    :param clim: Colormap limits. Default is None, in which case the min and
    max values of the array are used.
    :param cmap: Matplotlib colormap.
    :param figsize: Figure size in inches.
    :param dpi: Figure DPI.

    """

    def update(date):
        logging.debug("Plotting {}".format(date.strftime("%D")))
        image.set_data(da.sel(time=date))

        image_title.set_text("{:04d}/{:02d}/{:02d}".format(
            date.year, date.month, date.day))

        return image, image_title

    logging.info("Inspecting data")

    if clim is not None:
        n_min = clim[0]
        n_max = clim[1]
    else:
        n_max = da.max().values
        n_min = da.min().values

        if data_type == 'anom':
            if np.abs(n_max) > np.abs(n_min):
                n_min = -n_max
            elif np.abs(n_min) > np.abs(n_max):
                n_max = -n_min

    if video_dates is None:
        video_dates = [pd.Timestamp(date).to_pydatetime()
                       for date in da.time.values]

    if crop is not None:
        a = crop[0][0]
        b = crop[0][1]
        c = crop[1][0]
        d = crop[1][1]
        da = da.isel(xc=np.arange(a, b), yc=np.arange(c, d))
        if mask is not None:
            mask = mask[a:b, c:d]

    logging.info("Initialising plot")

    fig, ax = plt.subplots(figsize=(figsize, figsize))
    fig.set_dpi(dpi)

    if mask is not None:
        if mask_type == 'contour':
            ax.contour(mask, levels=[.5, 1], colors='k', zorder=3)
        elif mask_type == 'contourf':
            ax.contourf(mask, levels=[.5, 1], colors='k', zorder=3)

    ax.axes.xaxis.set_visible(False)
    ax.axes.yaxis.set_visible(False)

    divider = make_axes_locatable(ax)
    date = pd.Timestamp(da.time.values[0]).to_pydatetime()
    image = ax.imshow(da.sel(time=date),
                      cmap=cmap,
                      clim=(n_min, n_max),
                      animated=True,
                      zorder=1)

    image_title = ax.set_title("{:04d}/{:02d}/{:02d}".
                               format(date.year, date.month, date.day),
                               fontsize="medium", zorder=2)
    cax = divider.append_axes('right', size='5%', pad=0.05, zorder=2)
    plt.colorbar(image, cax)

    logging.info("Animating")
    # TODO: Investigate blit, but it causes a few problems with masks/titles.
    animation = FuncAnimation(fig,
                              update,
                              video_dates,
                              interval=1000/fps)

    plt.close()

    if not video_path:
        logging.info("Not saving plot, will return animation")
    else:
        logging.info("Saving plot to {}".format(video_path))
        animation.save(video_path,
                       fps=fps,
                       extra_args=['-vcodec', 'libx264'])
    return animation


def recurse_data_folders(base_path: object,
                         lookups: object,
                         children: object,
                         filetype: str = "nc") -> object:
    """

    :param base_path:
    :param lookups:
    :param children:
    :param filetype:
    :return:
    """
    logging.info("Looking at {}".format(base_path))
    files = []

    if children is None and lookups is None:
        # TODO: should ideally use scandir for performance
        # TODO: naive hardcoded filtering of files
        logging.debug("CHILDREN: {} or LOOKUPS: {}".format(children, lookups))
        files = sorted(
            [os.path.join(base_path, f) for f in os.listdir(base_path)
             if os.path.splitext(f)[1] == ".{}".format(filetype)
             and (re.match(r'^\d{4}\.nc$', f)
                  or
                  re.search(r'(abs|anom|linear_trend)\.nc$', f))])
        
        if not len(files):
            return None
    else:
        for subdir in os.listdir(base_path):
            logging.debug("SUBDIR: {}".format(subdir))
            new_path = os.path.join(base_path, subdir)

            if not os.path.isdir(new_path):
                continue

            if not len(lookups) or \
                    (len(lookups) and subdir in [str(s) for s in lookups]):
                subdir_files = recurse_data_folders(
                    new_path,
                    children[0]
                    if children is not None and len(children) > 0 else None,
                    children[1:]
                    if children is not None and len(children) > 1 else None,
                    filetype
                )
                if subdir_files:
                    files.append(subdir_files)

    return files


def video_process(files: object,
                  numpy: object,
                  output_dir: object,
                  fps: int) -> object:
    """

    :param files:
    :param numpy:
    :param output_dir:
    :param fps:
    :return:
    """
    path_comps = os.path.dirname(files[0]).split(os.sep)
    os.makedirs(output_dir, exist_ok=True)
    output_name = os.path.join(output_dir,
                               "{}.mp4".format("_".join(path_comps)))

    if not os.path.exists(output_name):
        logging.debug("Supplied: {} files for processing".format(len(files)))
        da = get_dataarray_from_files(files, numpy)
        logging.info("Saving to {}".format(output_name))
        xarray_to_video(da, fps, video_path=output_name)
    else:
        logging.warning("Not overwriting existing: {}".format(output_name))
        return None

    return output_name


@setup_logging
def cli_args():
    """

    :return:
    """
    args = argparse.ArgumentParser()

    args.add_argument("-f", "--fps", default=15, type=int)
    args.add_argument("-n", "--numpy", action="store_true", default=False)
    args.add_argument("-o", "--output-dir", dest="output_dir", type=str,
                      default="plot")
    args.add_argument("-p", "--path", default="data", type=str)
    args.add_argument("-w", "--workers", default=8, type=int)

    args.add_argument("-v", "--verbose", action="store_true", default=False)

    args.add_argument("data", type=lambda s: s.split(","))
    args.add_argument("hemisphere", default=[],
                      choices=["north", "south"], nargs="?")

    args.add_argument("--vars", default=[], type=lambda s: s.split(","))
    args.add_argument("--years", default=[], type=lambda s: s.split(","))

    return args.parse_args()


def data_cli():
    """

    """
    args = cli_args()

    hemis = [args.hemisphere] if len(args.hemisphere) else ["north", "south"]
    logging.info("Looking into {}".format(args.path))

    path_children = [hemis, args.vars]
    video_batches = recurse_data_folders(args.path,
                                         args.data,
                                         path_children,
                                         filetype="nc"
                                         if not args.numpy else "npy")
    video_batches = [
        v_el for h_list in video_batches
        for v_list in h_list
        for v_el in v_list
    ]

    if len(args.years) > 0:
        new_batches = []
        for batch in video_batches:
            batch = [el for el in batch
                     if os.path.basename(el)[0:4] in args.years]
            if len(batch):
                new_batches.append(batch)
            video_batches = new_batches

    logging.debug("Batches {}".format(video_batches))

    with ProcessPoolExecutor(
            max_workers=min(len(video_batches), args.workers)) as executor:
        futures = []

        for batch in video_batches:
            futures.append(executor.submit(video_process,
                                           batch,
                                           args.numpy,
                                           args.output_dir,
                                           args.fps))

        for future in as_completed(futures):
            try:
                res = future.result()

                if res:
                    logging.info("Produced {}".format(res))
            except Exception as e:
                logging.error(e)





import argparse
import datetime as dt
import logging
import os
import re

import numpy as np
import pandas as pd
import tensorflow as tf

import icenet.model.models as models

from icenet.data.dataset import IceNetDataSet
from icenet.utils import setup_logging

"""

"""


def predict_forecast(
    dataset_config: object,
    network_name: object,
    dataset_name: object = None,
    model_func: callable = models.unet_batchnorm,
    n_filters_factor: float = 1 / 8,
    network_folder: object = None,
    output_folder: object = None,
    seed: int = 42,
    start_dates: object = tuple([dt.datetime.now().date()]),
    test_set: bool = False,
) -> object:
    """

    :param dataset_config:
    :param network_name:
    :param dataset_name:
    :param model_func:
    :param n_filters_factor:
    :param network_folder:
    :param output_folder:
    :param seed:
    :param start_dates:
    :param test_set:
    :return:
    """
    # TODO: going to need to be able to handle merged datasets
    ds = IceNetDataSet(dataset_config)
    dl = ds.get_data_loader()

    if not network_folder:
        network_folder = os.path.join(".", "results", "networks", network_name)

    dataset_name = dataset_name if dataset_name else ds.identifier
    network_path = os.path.join(network_folder,
                                "{}.network_{}.{}.h5".format(network_name,
                                                             dataset_name,
                                                             seed))

    logging.info("Loading model from {}...".format(network_path))

    network = model_func(
        (*ds.shape, dl.num_channels),
        [],
        [],
        n_filters_factor=n_filters_factor,
        n_forecast_days=ds.n_forecast_days
    )
    network.load_weights(network_path)

    if not test_set:
        logging.info("Generating forecast inputs from processed/ files")

        for date in start_dates:
            run_prediction(network,
                           date,
                           output_folder,
                           *dl.generate_sample(date, prediction=True))
    else:
        # TODO: This is horrible behaviour, rethink and refactor: we should
        #  be able to pull from the test set in a nicer and more efficient
        #  fashion
        _, _, test_inputs = ds.get_split_datasets()

        source_key = [k for k in dl.config['sources'].keys() if k != "meta"][0]
        # FIXME: should be using date format from class
        test_dates = [dt.date(*[int(v) for v in d.split("_")]) for d in
                      dl.config["sources"][source_key]["dates"]["test"]]

        if len(test_dates) == 0:
            raise RuntimeError("No processed files were produced for the test "
                               "set")

        missing = set(start_dates).difference(test_dates)
        if len(missing) > 0:
            raise RuntimeError("{} are not in the test set".
                               format(", ".join([str(pd.to_datetime(el).date()) 
                                                 for el in missing])))

        data_iter = test_inputs.as_numpy_iterator()
        # FIXME: this is broken, this entry never gets added to the set?
        data = next(data_iter)
        x, y, sw = data
        batch = 0

        for i, idx in enumerate([test_dates.index(sd) for sd in start_dates]):
            while batch < int(idx / ds.batch_size):
                data = next(data_iter)
                x, y, sw = data
                batch += 1
            arr_idx = idx % ds.batch_size
            logging.info("Processing test batch {}, item {} (date {})".format(
                batch + 1, arr_idx, test_dates[idx]))

            run_prediction(network,
                           test_dates[idx],
                           output_folder,
                           x[arr_idx, ...],
                           y[arr_idx, ...],
                           sw[arr_idx, ...])


def run_prediction(network, date, output_folder,
                   net_input, net_output, sample_weights):
    logging.info("Running prediction {}".format(date))
    pred = network(tf.convert_to_tensor([net_input]), training=False)

    if os.path.exists(output_folder):
        logging.warning("{} output already exists".format(output_folder))
    os.makedirs(output_folder, exist_ok=output_folder)
    output_path = os.path.join(output_folder, date.strftime("%Y_%m_%d.npy"))

    logging.info("Saving {} - forecast output {}".format(date, pred.shape))
    np.save(output_path, pred)

    logging.debug("Saving loader generated data for reference...")

    for date, output, directory in \
            ((date, net_input, "input"),
             (date, net_output, "outputs"),
             (date, sample_weights, "weights")):
        output_directory = os.path.join(output_folder, "loader", directory)
        os.makedirs(output_directory, exist_ok=True)
        loader_output_path = os.path.join(output_directory,
                                          date.strftime("%Y_%m_%d.npy"))

        logging.info("Saving {} - generated {} {}".
                     format(date, directory, output.shape))
        np.save(loader_output_path, output)

    return output_path


def date_arg(string: str) -> object:
    """

    :param string:
    :return:
    """
    date_match = re.search(r"(\d{4})-(\d{1,2})-(\d{1,2})", string)
    return dt.date(*[int(s) for s in date_match.groups()])


@setup_logging
def get_args():
    """

    :return:
    """
    ap = argparse.ArgumentParser()
    ap.add_argument("dataset")
    ap.add_argument("network_name")
    ap.add_argument("output_name")
    ap.add_argument("seed", type=int, default=42)
    ap.add_argument("datefile", type=argparse.FileType("r"))

    ap.add_argument("-i", "--train-identifier", dest="ident",
                    help="Train dataset identifier", type=str, default=None)
    ap.add_argument("-n", "--n-filters-factor", type=float, default=1.)
    ap.add_argument("-t", "--testset", default=False, action="store_true")
    ap.add_argument("-v", "--verbose", action="store_true", default=False)

    return ap.parse_args()


def main():
    args = get_args()

    dataset_config = \
        os.path.join(".", "dataset_config.{}.json".format(args.dataset))

    date_content = args.datefile.read()
    dates = [dt.date(*[int(v) for v in s.split("-")])
             for s in date_content.split()]
    args.datefile.close()

    output_folder = os.path.join(".", "results", "predict",
                                 args.output_name,
                                 "{}.{}".format(args.network_name, args.seed))

    predict_forecast(dataset_config,
                     args.network_name,
                     # FIXME: this is turning into a mapping mess,
                     #  do we need to retain the train SD name in the
                     #  network?
                     dataset_name=
                     args.ident if args.ident else args.dataset,
                     n_filters_factor=
                     args.n_filters_factor,
                     output_folder=output_folder,
                     seed=args.seed,
                     start_dates=dates,
                     test_set=args.testset)




import tensorflow as tf
from tensorflow.keras import backend as K

"""
TensorFlow metrics.
"""


class ConstructLeadtimeAccuracy(tf.keras.metrics.CategoricalAccuracy):
    """

    Computes the network's accuracy over the active grid cell region
    for either a) a specific lead time in months, or b) over all lead times
    at once.

    :param name:
    :param use_all_forecast_months:
    :param single_forecast_leadtime_idx:
    """

    def __init__(self,
                 name: str = 'construct_custom_categorical_accuracy',
                 use_all_forecast_months: bool = True,
                 single_forecast_leadtime_idx: object = None,
                 **kwargs):
        super().__init__(name=name, **kwargs)

        self.use_all_forecast_months = use_all_forecast_months
        self.single_forecast_leadtime_idx = single_forecast_leadtime_idx

        raise NotImplementedError("Not updated for daily usage")

    def update_state(self,
                     y_true: object,
                     y_pred: object,
                     sample_weight: object = None):
        """

        :param y_true:
        :param y_pred:
        :param sample_weight:
        """
        if self.use_all_forecast_months:
            # Make class dimension final dimension for CategoricalAccuracy
            y_true = tf.transpose(y_true, [0, 1, 2, 4, 3])
            y_pred = tf.transpose(y_pred, [0, 1, 2, 4, 3])
            if sample_weight is not None:
                sample_weight = tf.transpose(sample_weight, [0, 1, 2, 4, 3])

            super().update_state(
                y_true, y_pred, sample_weight=sample_weight)

        elif not self.use_all_forecast_months:

            super().update_state(
                y_true[..., self.single_forecast_leadtime_idx],
                y_pred[..., self.single_forecast_leadtime_idx],
                sample_weight=sample_weight[..., self.single_forecast_leadtime_idx]>0)

    def result(self):
        """

        :return:
        """
        return 100 * super().result()

    def get_config(self):
        """For saving and loading networks with this custom metric.

        :return:
        """
        return {
            'single_forecast_leadtime_idx': self.single_forecast_leadtime_idx,
            'use_all_forecast_months': self.use_all_forecast_months,
        }

    @classmethod
    def from_config(cls, config: object):
        """For saving and loading networks with this custom metric.
        :param config:
        :return:
        """
        return cls(**config)


class WeightedBinaryAccuracy(tf.keras.metrics.BinaryAccuracy):
    """

    :param leadtime_idx:
    """

    def __init__(self,
                 leadtime_idx=None, **kwargs):
        name = 'binacc'

        # Leadtime to compute metric over - leave as None to use all lead times
        if leadtime_idx is not None:
            name += str(leadtime_idx+1)
        self._leadtime_idx = leadtime_idx

        super().__init__(name=name, **kwargs)

    def update_state(self,
                     y_true: object,
                     y_pred: object,
                     sample_weight: object = None):
        """ Custom keras loss/metric for binary accuracy in classifying SIC>15%

        :param y_true: Ground truth outputs
        :param y_pred: Network predictions
        :param sample_weight: Pixelwise mask weighting for metric summation

        :return: Root mean squared error of SIC (%) (float)
        """

        y_true = y_true > 0.15
        y_pred = y_pred > 0.15

        if self._leadtime_idx is not None:
            y_true = y_true[..., self._leadtime_idx, 0]
            y_pred = y_pred[..., self._leadtime_idx]
            if sample_weight is not None:
                sample_weight = sample_weight[..., self._leadtime_idx, 0]

        # TF automatically reduces along final dimension - include dummy axis
        if y_true.shape[-1] != 1:
            y_true = tf.expand_dims(y_true, axis=-1)
        if y_pred.shape[-1] != 1:
            y_pred = tf.expand_dims(y_pred, axis=-1)

        if sample_weight is not None and sample_weight.shape[-1] != 1:
            sample_weight = tf.expand_dims(sample_weight, axis=-1)

        return super().update_state(y_true, y_pred, sample_weight=sample_weight)

    def result(self):
        """

        :return:
        """
        return 100 * super().result()

    def get_config(self):
        """ For saving and loading networks with this custom metric.

        :return:
        """
        return {
            'leadtime_idx': self._leadtime_idx,
        }


class WeightedMAE(tf.keras.metrics.MeanAbsoluteError):
    """ Custom keras loss/metric for mean absolute error

    :param name:
    :param leadtime_idx:
    """

    def __init__(self,
                 name: str = 'mae',
                 leadtime_idx: object = None,
                 **kwargs):
        # Leadtime to compute metric over - leave as None to use all lead times
        if leadtime_idx is not None:
            name += str(leadtime_idx+1)
        self._leadtime_idx = leadtime_idx

        super().__init__(name=name, **kwargs)
    
    def update_state(self,
                     y_true: object,
                     y_pred: object,
                     sample_weight: object = None):
        """

        :param y_true:
        :param y_pred:
        :param sample_weight:
        :return:
        """

        if self._leadtime_idx is not None:
            y_true = y_true[..., self._leadtime_idx, 0]
            y_pred = y_pred[..., self._leadtime_idx]
            if sample_weight is not None:
                sample_weight = sample_weight[..., self._leadtime_idx, 0]

        # TF automatically reduces along final dimension - include dummy axis
        if y_true.shape[-1] != 1:
            y_true = tf.expand_dims(y_true, axis=-1)
        if y_pred.shape[-1] != 1:
            y_pred = tf.expand_dims(y_pred, axis=-1)

        if sample_weight is not None and sample_weight.shape[-1] != 1:
            sample_weight = tf.expand_dims(sample_weight, axis=-1)

        return super().update_state(y_true, y_pred, sample_weight=sample_weight)

    def result(self):
        """

        :return:
        """
        return 100 * super().result()


class WeightedRMSE(tf.keras.metrics.RootMeanSquaredError):
    """ Custom keras loss/metric for root mean squared error

    :param leadtime_idx:
    :param name:
    """

    def __init__(self,
                 leadtime_idx: object = None,
                 name: str = 'rmse',
                 **kwargs):
        # Leadtime to compute metric over - leave as None to use all lead times
        if leadtime_idx is not None:
            name += str(leadtime_idx+1)
        self._leadtime_idx = leadtime_idx

        super().__init__(name=name, **kwargs)

    def update_state(self,
                     y_true: object,
                     y_pred: object,
                     sample_weight: object = None):
        """

        :param y_true:
        :param y_pred:
        :param sample_weight:
        :return:
        """
        if self._leadtime_idx is not None:
            y_true = y_true[..., self._leadtime_idx, 0]
            y_pred = y_pred[..., self._leadtime_idx]
            if sample_weight is not None:
                sample_weight = sample_weight[..., self._leadtime_idx, 0]

        # TF automatically reduces along final dimension - include dummy axis
        if y_true.shape[-1] != 1:
            y_true = tf.expand_dims(y_true, axis=-1)
        if y_pred.shape[-1] != 1:
            y_pred = tf.expand_dims(y_pred, axis=-1)

        if sample_weight is not None and sample_weight.shape[-1] != 1:
            sample_weight = tf.expand_dims(sample_weight, axis=-1)

        return super().update_state(y_true, y_pred, sample_weight=sample_weight)

    def result(self):
        """

        :return:
        """
        return 100 * super().result()


class WeightedMSE(tf.keras.metrics.MeanSquaredError):
    """ Custom keras loss/metric for mean squared error

    :param leadtime_idx:
    :param name:
    """

    def __init__(self,
                 leadtime_idx: object = None,
                 **kwargs):
        name = 'mse'
        # Leadtime to compute metric over - leave as None to use all lead times
        if leadtime_idx is not None:
            name += str(leadtime_idx+1)
        self._leadtime_idx = leadtime_idx

        super().__init__(name=name, **kwargs)

    def update_state(self,
                     y_true: object,
                     y_pred: object,
                     sample_weight: object = None):
        """

        :param y_true:
        :param y_pred:
        :param sample_weight:
        :return:
        """

        if self._leadtime_idx is not None:
            y_true = y_true[..., self._leadtime_idx, 0]
            y_pred = y_pred[..., self._leadtime_idx]
            if sample_weight is not None:
                sample_weight = sample_weight[..., self._leadtime_idx, 0]

        # TF automatically reduces along final dimension - include dummy axis
        if y_true.shape[-1] != 1:
            y_true = tf.expand_dims(y_true, axis=-1)
        if y_pred.shape[-1] != 1:
            y_pred = tf.expand_dims(y_pred, axis=-1)

        if sample_weight is not None and sample_weight.shape[-1] != 1:
            sample_weight = tf.expand_dims(sample_weight, axis=-1)

        return super().update_state(y_true, y_pred, sample_weight=sample_weight)

    def result(self):
        """

        :return:
        """
        return 100 * super().result()




from icenet.data.loaders.base import IceNetBaseDataLoader

"""
Python Standard Library implementations for icenet data loading

Still WIP to re-introduce alternate implementations that might work better in 
certain deployments

"""


class IceNetDataLoader(IceNetBaseDataLoader):
    def __init__(self,
                 *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        # TODO: https://github.com/icenet-ai/icenet/blob/cb68e5dec31d4c62d72411cbca4c6d3a0276e0f9/icenet2/data/loader.py
        raise NotImplementedError("Not yet adapted from old implementation")

    def generate(self):
        """

        """
        pass

    def generate_sample(self,
                        date: object,
                        prediction: bool = False):
        """

        :param date:
        :param prediction:
        """
        pass



import argparse
import datetime as dt
import logging

import iris
import pandas as pd
import xarray as xr

import iris.analysis

from icenet.process.utils import date_arg
from icenet.utils import setup_logging


def broadcast_forecast(start_date: object,
                       end_date: object,
                       datafiles: object = None,
                       dataset: object = None,
                       target: object = None) -> object:
    """

    :param start_date:
    :param end_date:
    :param datafiles:
    :param dataset:
    :param target:
    :return:
    """

    assert (datafiles is None) ^ (dataset is None), \
        "Only one of datafiles and dataset can be set"

    if datafiles:
        logging.info("Using {} to generate forecast through {} to {}".
                     format(", ".join(datafiles), start_date, end_date))
        dataset = xr.open_mfdataset(datafiles, engine="netcdf4")

    dates = pd.date_range(start_date, end_date)
    i = 0

    logging.debug("Dataset summary: \n{}".format(dataset))

    if len(dataset.time.values) > 1:
        while dataset.time.values[i + 1] < dates[0]:
            i += 1

    logging.info("Starting index will be {} for {} - {}".
                 format(i, dates[0], dates[-1]))
    dt_arr = []

    for d in dates:
        logging.debug("Looking for date {}".format(d))
        arr = None

        while arr is None:
            if d >= dataset.time.values[i]:
                d_lead = (d - dataset.time.values[i]).days

                if i + 1 < len(dataset.time.values):
                    if pd.to_datetime(dataset.time.values[i]) + \
                            dt.timedelta(days=d_lead) >= \
                            pd.to_datetime(dataset.time.values[i + 1]) + \
                            dt.timedelta(days=1):
                        i += 1
                        continue

                logging.debug("Selecting date {} and lead {}".
                              format(pd.to_datetime(
                                     dataset.time.values[i]).strftime("%D"), 
                                     d_lead))

                arr = dataset.sel(time=dataset.time.values[i],
                                  leadtime=d_lead).\
                    copy().\
                    drop("time").\
                    assign_coords(dict(time=d)).\
                    drop("leadtime")
            else:
                i += 1

        dt_arr.append(arr)

    target_ds = xr.concat(dt_arr, dim="time")

    if target:
        logging.info("Saving dataset to {}".format(target))
        target_ds.to_netcdf(target)
    return target_ds


def reproject_output(forecast_file: object,
                     proj_file: object,
                     save_file: object) -> object:
    """

    :param forecast_file:
    :param proj_file:
    :param save_file:
    """
    logging.info("Loading forecast {}".format(forecast_file))
    forecast_cube = iris.load_cube(forecast_file)

    logging.info("Projecting as per {}".format(proj_file))
    gp = iris.load_cube(proj_file)

    forecast_cube.coord('projection_y_coordinate').convert_units('meters')
    forecast_cube.coord('projection_x_coordinate').convert_units('meters')

    logging.info("Attempting to reproject and save to {}".
                 format(save_file))
    latlon_cube = forecast_cube.regrid(gp, iris.analysis.Linear())
    iris.save(latlon_cube, save_file)


@setup_logging
def broadcast_args():
    ap = argparse.ArgumentParser()
    ap.add_argument("-o", "--output-target", dest="target", default=None)
    ap.add_argument("start_date", type=date_arg)
    ap.add_argument("end_date", type=date_arg)
    ap.add_argument("datafiles", nargs="+")
    args = ap.parse_args()
    return args


def broadcast_main():
    args = broadcast_args()
    broadcast_forecast(args.start_date, args.end_date,
                       args.datafiles, args.target)


@setup_logging
def reproject_args():
    ap = argparse.ArgumentParser()
    ap.add_argument("forecast_file")
    ap.add_argument("proj_file")
    ap.add_argument("save_file")
    args = ap.parse_args()
    return args


def reproject_main():
    args = reproject_args()
    reproject_output(args.forecast_file, args.proj_file, args.save_file)



import logging
import os
import warnings

import numpy as np
import pandas as pd
import xarray as xr

from icenet.data.interfaces.downloader import ClimateDownloader
from icenet.data.cli import download_args
from icenet.data.utils import esgf_search

"""

"""


class CMIP6Downloader(ClimateDownloader):
    """Climate downloader to provide CMIP6 reanalysis data from ESGF APIs

    Useful CMIP6 guidance: https://pcmdi.llnl.gov/CMIP6/Guide/dataUsers.html

    :param identifier: how to identify this dataset
    :param source: source ID in ESGF node
    :param member: member ID in ESGF node
    :param nodes: list of ESGF nodes to query
    :param experiments: experiment IDs to download
    :param frequency: query parameter frequency
    :param table_map: table map for
    :param grid_map:
    :param grid_override:
    :param exclude_nodes:

    "MRI-ESM2-0", "r1i1p1f1", None
    "MRI-ESM2-0", "r2i1p1f1", None
    "MRI-ESM2-0", "r3i1p1f1", None
    "MRI-ESM2-0", "r4i1p1f1", None
    "MRI-ESM2-0", "r5i1p1f1", None
    "EC-Earth3", "r2i1p1f1", "gr"
    "EC-Earth3", "r7i1p1f1", "gr"
    "EC-Earth3", "r10i1p1f1", "gr"
    "EC-Earth3", "r12i1p1f1", "gr"
    "EC-Earth3", "r14i1p1f1", "gr"

    """
    
    TABLE_MAP = {
        'siconca': 'SIday',
        'tas': 'day',
        'ta': 'day',
        'tos': 'Oday',
        'hus': 'day',
        'psl': 'day',
        'rlds': 'day',
        'rsus': 'day',
        'rsds': 'day',
        'zg': 'day',
        'uas': 'day',
        'vas': 'day',
        'ua': 'day',
    }

    GRID_MAP = {
        'siconca': 'gn',
        'tas': 'gn',
        'ta': 'gn',
        'tos': 'gr',
        'hus': 'gn',
        'psl': 'gn',
        'rlds': 'gn',
        'rsus': 'gn',   # Surface Upwelling Shortwave Radiation
        'rsds': 'gn',   # Surface Downwelling Shortwave Radiation
        'zg': 'gn',
        'uas': 'gn',
        'vas': 'gn',
        'ua': 'gn',
    }

    # Prioritise European first, US last, avoiding unnecessary queries
    # against nodes further afield (all traffic has a cost, and the coverage
    # of local nodes is more than enough)
    ESGF_NODES = ("esgf.ceda.ac.uk",
                  "esg1.umr-cnrm.fr",
                  "vesg.ipsl.upmc.fr",
                  "esgf3.dkrz.de",
                  "esgf.bsc.es",
                  "esgf-data.csc.fi",
                  "noresg.nird.sigma2.no",
                  "esgf-data.ucar.edu",
                  "esgf-data2.diasjp.net",)

    def __init__(self,
                 *args,
                 source: str,
                 member: str,
                 nodes: object = ESGF_NODES,
                 experiments: object = ('historical', 'ssp245'),
                 frequency: str = "day",
                 table_map: object = None,
                 grid_map: object = None,
                 grid_override: object = None,
                 exclude_nodes: object = None,
                 **kwargs):
        super().__init__(*args,
                         identifier="cmip6.{}.{}".format(source, member),
                         **kwargs)

        self._source = source
        self._member = member
        self._frequency = frequency
        self._experiments = experiments

        self._nodes = nodes if not exclude_nodes else \
            [n for n in nodes if n not in exclude_nodes]

        self._table_map = table_map if table_map else CMIP6Downloader.TABLE_MAP
        self._grid_map = grid_map if grid_map else CMIP6Downloader.GRID_MAP
        self._grid_map_override = grid_override

    def _single_download(self,
                         var_prefix: str,
                         level: object,
                         req_dates: object):
        """Overridden CMIP implementation for downloading from DAP server

        Due to the size of the CMIP set and the fact that we don't want to make
        1850-2100 yearly requests for all downloads, we have a bespoke and
        overridden download implementation for this.

        :param var_prefix:
        :param level:
        :param req_dates:
        """

        query = {
            'source_id': self._source,
            'member_id': self._member,
            'frequency': self._frequency,
            'variable_id': var_prefix,
            'table_id': self._table_map[var_prefix],
            'grid_label': self._grid_map_override
            if self._grid_map_override
            else self._grid_map[var_prefix],
        }

        var = var_prefix if not level else "{}{}".format(var_prefix, level)

        req_dates, merge_files = \
            self.filter_dates_on_data(var_prefix, level, req_dates)

        logging.info("Querying ESGF")
        results = []

        for experiment_id in self._experiments:
            query['experiment_id'] = experiment_id

            for data_node in self._nodes:
                query['data_node'] = data_node

                # FIXME: inefficient, we can strip redundant results files
                #  based on WCRP data management standards for file naming,
                #  such as based on date. Refactor/rewrite this impl...
                node_results = esgf_search(**query)

                if len(node_results):
                    logging.debug("Query: {}".format(query))
                    logging.debug("Found {}: {}".format(experiment_id,
                                                        node_results))
                    results.extend(node_results)
                    break

        logging.info("Found {} {} results from ESGF search".
                     format(len(results), var_prefix))

        try:
            # http://xarray.pydata.org/en/stable/user-guide/io.html?highlight=opendap#opendap
            # Avoid 500MB DAP request limit
            cmip6_da = xr.open_mfdataset(results,
                                         combine='by_coords',
                                         chunks={'time': '499MB'}
                                         )[var_prefix]

            cmip6_da = cmip6_da.sel(time=slice(req_dates[0],
                                               req_dates[-1]))

            # TODO: possibly other attributes, especially with ocean vars
            if level:
                cmip6_da = cmip6_da.sel(plev=int(level) * 100)

            cmip6_da = cmip6_da.sel(lat=slice(self.hemisphere_loc[2],
                                              self.hemisphere_loc[0]))
            self.save_temporal_files(var, cmip6_da)
        except OSError as e:
            logging.exception("Error encountered: {}".format(e),
                              exc_info=False)

    def additional_regrid_processing(self,
                                     datafile: str,
                                     cube_ease: object):
        """

        :param datafile:
        :param cube_ease:
        """
        (datafile_path, datafile_name) = os.path.split(datafile)
        var_name = datafile_path.split(os.sep)[self._var_name_idx]

        # TODO: regrid fixes need better implementations
        if var_name == "siconca":
            cube_ease.data[cube_ease.data.mask] = 0.
            cube_ease.data[:, self._masks.get_land_mask()] = 0.

            if self._source == 'MRI-ESM2-0':
                cube_ease.data = cube_ease.data / 100.
            cube_ease.data = cube_ease.data.data
        elif var_name in ["tos", "hus1000"]:
            cube_ease.data[cube_ease.data.mask] = 0.
            cube_ease.data[:, self._masks.get_land_mask()] = 0.

            cube_ease.data = cube_ease.data.data

        if cube_ease.data.dtype != np.float32:
            logging.info("Regrid processing, data type not float: {}".
                         format(cube_ease.data.dtype))
            cube_ease.data = cube_ease.data.astype(np.float32)

    def convert_cube(self, cube: object) -> object:
        """Converts Iris cube to be fit for CMIP regrid

        :param cube:   the cube requiring alteration
        :return cube:   the altered cube
        """

        cs = self.sic_ease_cube.coord_system().ellipsoid

        for coord in ['longitude', 'latitude']:
            cube.coord(coord).coord_system = cs
        return cube


def main():
    args = download_args(
        dates=True,
        extra_args=[
            (["source"], dict(type=str)),
            (["member"], dict(type=str)),
            (("-xs", "--exclude-server"),
             dict(default=[], nargs="*")),
            (("-o", "--override"), dict(required=None, type=str)),
        ],
        workers=True
    )

    logging.info("CMIP6 Data Downloading")

    downloader = CMIP6Downloader(
        source=args.source,
        member=args.member,
        var_names=args.vars,
        pressure_levels=args.levels,
        dates=[pd.to_datetime(date).date() for date in
               pd.date_range(args.start_date, args.end_date, freq="D")],
        delete_tempfiles=args.delete,
        grid_override=args.override,
        north=args.hemisphere == "north",
        south=args.hemisphere == "south",
        max_threads=args.workers,
        exclude_nodes=args.exclude_server,
    )
    logging.info("CMIP downloading: {} {}".format(args.source, args.member))
    downloader.download()

    logging.info("CMIP regridding: {} {}".format(args.source, args.member))
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=UserWarning)
        downloader.regrid()



from setuptools import setup, find_packages

import icenet

"""Setup module for icenet
"""


def get_content(filename):
    with open(filename, "r") as fh:
        return fh.read()


setup(
    name=icenet.__name__,
    version=icenet.__version__,
    author=icenet.__author__,
    author_email=icenet.__email__,
    description="Library for operational IceNet forecasting",
    long_description="""{}\n---\n""".
                     format(get_content("README.md"),
                            get_content("HISTORY.rst")),
    long_description_content_type="text/markdown",
    url="https://github.com/icenet-ai",
    packages=find_packages(),
    keywords="",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    entry_points={
        "console_scripts": [
            "icenet_data_masks = icenet.data.sic.mask:main",

            "icenet_data_cmip = icenet.data.interfaces.esgf:main",
            "icenet_data_era5 = icenet.data.interfaces.cds:main",
            "icenet_data_oras5 = icenet.data.interfaces.cmems:main",
            "icenet_data_hres = icenet.data.interfaces.mars:hres_main",
            "icenet_data_seas = icenet.data.interfaces.mars:seas_main",
            "icenet_data_sic = icenet.data.sic.osisaf:main",

            "icenet_data_reproc_monthly = "
            "icenet.data.interfaces.utils:reprocess_main",
            "icenet_data_add_time_dim = "
            "icenet.data.interfaces.utils:add_time_dim_main",

            "icenet_process_cmip = icenet.data.processors.cmip:main",
            "icenet_process_era5 = icenet.data.processors.era5:main",
            "icenet_process_oras5 = icenet.data.processors.oras5:main",
            "icenet_process_hres = icenet.data.processors.hres:main",
            "icenet_process_sic = icenet.data.processors.osi:main",

            "icenet_process_metadata = icenet.data.processors.meta:main",

            "icenet_process_condense = "
            "icenet.data.processors.utils:condense_main",

            "icenet_dataset_check = icenet.data.dataset:check_dataset",
            "icenet_dataset_create = icenet.data.loader:create",
            "icenet_loader_sample = icenet.data.loader:get_sample",

            "icenet_train = icenet.model.train:main",
            "icenet_predict = icenet.model.predict:main",
            "icenet_upload_azure = icenet.process.azure:upload",
            "icenet_upload_local = icenet.process.local:upload",

            "icenet_plot_sample = icenet.plotting.data:plot_sample",
            "icenet_plot_record = icenet.plotting.data:plot_tfrecord",

            "icenet_plot_sic_error = icenet.plotting.forecast:sic_error",
            "icenet_plot_bin_accuracy = "
            "icenet.plotting.forecast:binary_accuracy",

            "icenet_video_data = icenet.plotting.video:data_cli",

            "icenet_output = icenet.process.predict:create_cf_output",
            "icenet_output_broadcast = "
            "icenet.process.forecasts:broadcast_main",
            "icenet_output_reproject = "
            "icenet.process.forecasts:reproject_main",
        ],
    },
    python_requires='>=3.7, <4',
    install_requires=get_content("requirements.txt"),
    include_package_data=True,
    extras_require={
        "dev": get_content("requirements_dev.txt"),
        "docs": get_content("docs/requirements.txt"),
    },
    test_suite='tests',
    tests_require=['pytest>=3'],
    zip_safe=False,
)



import argparse
import datetime as dt
import json
import logging
import os
import random
import time

from pprint import pformat

import numpy as np
import pandas as pd
import tensorflow as tf
import wandb

from tensorflow.keras.callbacks import \
    EarlyStopping, ModelCheckpoint, LearningRateScheduler
from tensorflow.keras.models import load_model, save_model
from wandb.keras import WandbCallback

from icenet.data.dataset import IceNetDataSet, MergedIceNetDataSet
import icenet.model.losses as losses
import icenet.model.metrics as metrics
from icenet.model.utils import make_exp_decay_lr_schedule
import icenet.model.models as models
from icenet.utils import setup_logging


def train_model(
        run_name: object,
        dataset: object,
        batch_size: int = 4,
        checkpoint_monitor: str = 'val_rmse',
        checkpoint_mode: str = 'min',
        dataset_ratio: float = 1.0,
        early_stopping_patience: int = 30,
        epochs: int = 2,
        filter_size: float = 3,
        learning_rate: float = 1e-4,
        lr_10e_decay_fac: float = 1.0,
        lr_decay_start: float = 10,
        lr_decay_end: float = 30,
        max_queue_size: int = 3,
        model_func: object = models.unet_batchnorm,
        n_filters_factor: float = 2,
        network_folder: object = None,
        network_save: bool = True,
        pickup_weights: bool = False,
        pre_load_network: bool = False,
        pre_load_path: object = None,
        seed: int = 42,
        strategy: object = tf.distribute.get_strategy(),
        training_verbosity: int = 1,
        workers: int = 5,
        use_multiprocessing: bool = True,
        use_tensorboard: bool = True,
        use_wandb: bool = True,
        wandb_offline: bool = False,
        wandb_project: str = os.environ.get("ICENET_ENVIRONMENT"),
        wandb_user: str = os.environ.get("USER")) -> object:
    """

    :param run_name:
    :param dataset:
    :param batch_size:
    :param checkpoint_monitor:
    :param checkpoint_mode:
    :param dataset_ratio:
    :param early_stopping_patience:
    :param epochs:
    :param filter_size:
    :param learning_rate:
    :param lr_10e_decay_fac:
    :param lr_decay_start:
    :param lr_decay_end:
    :param max_queue_size:
    :param model_func:
    :param n_filters_factor:
    :param network_folder:
    :param network_save:
    :param pickup_weights:
    :param pre_load_network:
    :param pre_load_path:
    :param seed:
    :param strategy:
    :param training_verbosity:
    :param workers:
    :param use_multiprocessing:
    :param use_tensorboard:
    :param use_wandb:
    :param wandb_offline:
    :param wandb_project:
    :param wandb_user:
    :return:
    """

    lr_decay = -0.1 * np.log(lr_10e_decay_fac)
    wandb.init(
        project=wandb_project,
        name="{}.{}".format(run_name, seed),
        notes="{}: run at {}{}".format(run_name,
                                       dt.datetime.now().strftime("%D %T"),
                                       "" if
                                       not pre_load_network else
                                       " preload {}".format(pre_load_path)),
        entity=wandb_user,
        config=dict(
            seed=seed,
            learning_rate=learning_rate,
            filter_size=filter_size,
            n_filters_factor=n_filters_factor,
            lr_10e_decay_fac=lr_10e_decay_fac,
            lr_decay=lr_decay,
            lr_decay_start=lr_decay_start,
            lr_decay_end=lr_decay_end,
            batch_size=batch_size,
        ),
        allow_val_change=True,
        mode='disabled' if not use_wandb else 'offline' if wandb_offline else 'online',
        settings=wandb.Settings(
            start_method="fork",
            _disable_stats=True,
        ),
        group=run_name,
    )

    logging.info("Hyperparameters: {}".format(pformat(wandb.config)))

    input_shape = (*dataset.shape, dataset.num_channels)

    if pre_load_network and not os.path.exists(pre_load_path):
        raise RuntimeError("{} is not available, so you cannot preload the "
                           "network with it!".format(pre_load_path))

    if not network_folder:
        network_folder = os.path.join(".", "results", "networks", run_name)

    if not os.path.exists(network_folder):
        logging.info("Creating network folder: {}".format(network_folder))
        os.makedirs(network_folder, exist_ok=True)

    weights_path = os.path.join(network_folder,
                                "{}.network_{}.{}.h5".format(run_name,
                                                             dataset.identifier,
                                                             seed))
    model_path = os.path.join(network_folder,
                              "{}.model_{}.{}".format(run_name,
                                                      dataset.identifier,
                                                      seed))

    history_path = os.path.join(network_folder,
                                "{}_{}_history.json".format(run_name, seed))

    prev_best = None
    callbacks_list = list()

    # Checkpoint the model weights when a validation metric is improved
    callbacks_list.append(
        ModelCheckpoint(
            filepath=weights_path,
            monitor=checkpoint_monitor,
            verbose=1,
            mode=checkpoint_mode,
            save_best_only=True
        ))

    # Abort training when validation performance stops improving
    callbacks_list.append(
        EarlyStopping(
            monitor=checkpoint_monitor,
            mode=checkpoint_mode,
            verbose=1,
            patience=early_stopping_patience,
            baseline=prev_best
        ))

    callbacks_list.append(
        LearningRateScheduler(
            make_exp_decay_lr_schedule(
                rate=lr_decay,
                start_epoch=lr_decay_start,
                end_epoch=lr_decay_end,
            )))

    if use_tensorboard:
        logging.info("Adding tensorboard callback")
        log_dir = "logs/" + dt.datetime.now().strftime("%d-%m-%y-%H%M%S")
        callbacks_list.append(tf.keras.callbacks.TensorBoard(log_dir=log_dir,
                                                             histogram_freq=1))

    if use_wandb:
        # Log training metrics to wandb each epoch
        logging.info("Adding wandb callback")
        callbacks_list.append(
            WandbCallback(
                monitor=checkpoint_monitor,
                mode=checkpoint_mode,
                save_model=False,
                save_graph=False,
            ))

    ############################################################################
    #                              TRAINING MODEL
    ############################################################################

    with strategy.scope():
        loss = losses.WeightedMSE()
        metrics_list = [
            # metrics.weighted_MAE,
            metrics.WeightedBinaryAccuracy(),
            metrics.WeightedMAE(),
            metrics.WeightedRMSE(),
            losses.WeightedMSE()
        ]

        network = model_func(
            input_shape=input_shape,
            loss=loss,
            metrics=metrics_list,
            learning_rate=learning_rate,
            filter_size=filter_size,
            n_filters_factor=n_filters_factor,
            n_forecast_days=dataset.n_forecast_days,
        )

    if pre_load_network:
        logging.info("Loading network weights from {}".format(pre_load_path))
        network.load_weights(pre_load_path)
    elif pickup_weights and os.path.exists(weights_path):
        logging.warning("Automagically loading network weights from {}".
                        format(weights_path))
        network.load_weights(weights_path)

    network.summary()

    ratio = dataset_ratio if dataset_ratio else 1.0
    train_ds, val_ds, test_ds = dataset.get_split_datasets(ratio=ratio)

    model_history = network.fit(
        train_ds,
        epochs=epochs,
        verbose=training_verbosity,
        callbacks=callbacks_list,
        validation_data=val_ds,
        max_queue_size=max_queue_size,
        # TODO: not useful for tf.data usage according to docs
        workers=workers,
        use_multiprocessing=use_multiprocessing
    )

    if network_save:
        logging.info("Saving network to: {}".format(weights_path))
        network.save_weights(weights_path)
        save_model(network, model_path)

        with open(history_path, 'w') as fh:
            pd.DataFrame(model_history.history).to_json(fh)

    return weights_path, model_path


def evaluate_model(model_path: object,
                   dataset: object,
                   dataset_ratio: float = 1.0,
                   max_queue_size: int = 3,
                   workers: int = 5,
                   use_multiprocessing: bool = True):
    """

    :param model_path:
    :param dataset:
    :param dataset_ratio:
    :param max_queue_size:
    :param workers:
    :param use_multiprocessing:
    """
    logging.info("Running evaluation against test set")
    network = load_model(model_path, compile=False)

    _, val_ds, test_ds = dataset.get_split_datasets(ratio=dataset_ratio)
    eval_data = val_ds

    if dataset.counts["test"] > 0:
        logging.warning("Using validation data source for evaluation, rather "
                        "than test set")
        eval_data = test_ds

    lead_times = list(range(1, dataset.n_forecast_days + 1))
    logging.info("Metric creation for lead time of {} days".
                 format(len(lead_times)))
    metric_names = ["binacc", "mae", "rmse"]
    metrics_classes = [
        metrics.WeightedBinaryAccuracy,
        metrics.WeightedMAE,
        metrics.WeightedRMSE,
    ]
    metrics_list = [cls(leadtime_idx=lt - 1)
                    for lt in lead_times
                    for cls in metrics_classes]

    network.compile(weighted_metrics=metrics_list)

    logging.info('Evaluating... ')
    tic = time.time()
    results = network.evaluate(
        eval_data,
        return_dict=True,
        verbose=0,
        max_queue_size=max_queue_size,
        workers=workers,
        use_multiprocessing=use_multiprocessing,
    )

    results_path = "{}.results.json".format(model_path)
    with open(results_path, "w") as fh:
        json.dump(results, fh)

    logging.debug(results)
    logging.info("Done in {:.1f}s".format(time.time() - tic))

    metric_vals = [[results[f'{name}{lt}']
                    for lt in lead_times] for name in metric_names]
    table_data = list(zip(lead_times, *metric_vals))
    table = wandb.Table(data=table_data, columns=['leadtime', *metric_names])

    # Log each metric vs. leadtime as a plot to wandb
    for name in metric_names:
        wandb.log(
            {f'{name}_plot': wandb.plot.line(table, x='leadtime', y=name)})


@setup_logging
def get_args():
    """

    :return:
    """
    ap = argparse.ArgumentParser()
    ap.add_argument("dataset", type=str)
    ap.add_argument("run_name", type=str)
    ap.add_argument("seed", type=int)

    ap.add_argument("-b", "--batch-size", type=int, default=4)
    ap.add_argument("-ds", "--additional-dataset",
                    dest="additional", nargs="*", default=[])
    ap.add_argument("-e", "--epochs", type=int, default=4)
    ap.add_argument("--early-stopping", type=int, default=50)
    ap.add_argument("-m", "--multiprocessing",
                    action="store_true", default=False)
    ap.add_argument("-n", "--n-filters-factor", type=float, default=1.)
    ap.add_argument("-nw", "--no-wandb", default=False, action="store_true")
    ap.add_argument("-p", "--preload", type=str)
    ap.add_argument("-pw", "--pickup-weights",
                    action="store_true", default=False)
    ap.add_argument("-qs", "--max-queue-size", default=10, type=int)
    ap.add_argument("-r", "--ratio", default=1.0, type=float)
    ap.add_argument("-s", "--strategy", default="default",
                    choices=("default", "mirrored", "central"))
    ap.add_argument("--shuffle-train", default=False,
                    action="store_true", help="Shuffle the training set")
    ap.add_argument("--gpus", default=None)
    ap.add_argument("-v", "--verbose", action="store_true", default=False)
    ap.add_argument("-w", "--workers", type=int, default=4)
    ap.add_argument("-wo", "--wandb-offline", default=False, action="store_true")
    ap.add_argument("-wp", "--wandb-project",
                    default=os.environ.get("ICENET_ENVIRONMENT"), type=str)
    ap.add_argument("-wu", "--wandb-user",
                    default=os.environ.get("USER"), type=str)

    ap.add_argument("--lr", default=1e-4, type=float)
    ap.add_argument("--lr_10e_decay_fac", default=1.0, type=float,
                    help="Factor by which LR is multiplied by every 10 epochs "
                         "using exponential decay. E.g. 1 -> no decay (default)"
                         ", 0.5 -> halve every 10 epochs.")
    ap.add_argument('--lr_decay_start', default=10, type=int)
    ap.add_argument('--lr_decay_end', default=30, type=int)

    return ap.parse_args()


def main():
    args = get_args()

    logging.warning("Setting seed for best attempt at determinism, value {}".
                    format(args.seed))
    # determinism is not guaranteed across different versions of TensorFlow.
    # determinism is not guaranteed across different hardware.
    os.environ['PYTHONHASHSEED'] = str(args.seed)
    # numpy.random.default_rng ignores this, WARNING!
    np.random.seed(args.seed)
    random.seed(args.seed)
    tf.random.set_seed(args.seed)
    tf.keras.utils.set_random_seed(args.seed)
    # See #8: tf.config.experimental.enable_op_determinism()

    # TODO: this should come from a factory in the future - not the only place
    #  that merged datasets are going to be available
    if len(args.additional) == 0:
        dataset = IceNetDataSet("dataset_config.{}.json".format(args.dataset),
                                batch_size=args.batch_size,
                                shuffling=args.shuffle_train)
    else:
        dataset = MergedIceNetDataSet([
            "dataset_config.{}.json".format(el) for el in [
                args.dataset, *args.additional
            ]
        ],
            batch_size=args.batch_size,
            shuffling=args.shuffle_train)

    strategy = tf.distribute.MirroredStrategy() \
        if args.strategy == "mirrored" \
        else tf.distribute.experimental.CentralStorageStrategy() \
        if args.strategy == "central" \
        else tf.distribute.get_strategy()

    weights_path, model_path = \
        train_model(args.run_name,
                    dataset,
                    batch_size=args.batch_size,
                    dataset_ratio=args.ratio,
                    early_stopping_patience=args.early_stopping,
                    epochs=args.epochs,
                    learning_rate=args.lr,
                    lr_10e_decay_fac=args.lr_10e_decay_fac,
                    lr_decay_start=args.lr_decay_start,
                    lr_decay_end=args.lr_decay_end,
                    pickup_weights=args.pickup_weights,
                    pre_load_network=args.preload is not None,
                    pre_load_path=args.preload,
                    max_queue_size=args.max_queue_size,
                    n_filters_factor=args.n_filters_factor,
                    seed=args.seed,
                    strategy=strategy,
                    training_verbosity=1 if args.verbose else 2,
                    use_multiprocessing=args.multiprocessing,
                    use_wandb=not args.no_wandb,
                    wandb_offline=args.wandb_offline,
                    wandb_project=args.wandb_project,
                    wandb_user=args.wandb_user,
                    workers=args.workers)

    evaluate_model(model_path,
                   dataset,
                   dataset_ratio=args.ratio,
                   max_queue_size=args.max_queue_size,
                   use_multiprocessing=args.multiprocessing,
                   workers=args.workers)



import argparse
import json
import logging
import os
import subprocess

import numpy as np
import tensorflow as tf

from icenet.data.datasets.utils import get_decoder
from icenet.utils import setup_logging

import matplotlib.pyplot as plt


@setup_logging
def tfrecord_args():
    """

    :return:
    """
    ap = argparse.ArgumentParser()
    ap.add_argument("file")
    ap.add_argument("configuration", type=argparse.FileType("r"))
    ap.add_argument("-i", "--index", default=1, type=int)
    ap.add_argument("-l", "--levels", default=100, type=int)
    ap.add_argument("-o", "--output", default="plot")

    return ap.parse_args()


def plot_tfrecord():
    args = tfrecord_args()

    ds = tf.data.TFRecordDataset([args.file])
    config = json.load(args.configuration)
    args.configuration.close()

    decoder = get_decoder(tuple(config['shape']),
                          config['num_channels'],
                          config['n_forecast_days'])

    ds = ds.map(decoder).batch(1)
    it = ds.as_numpy_iterator()

    for _ in range(0, args.index):
        data = next(it)

    x, y, sample_weights = data
    logging.debug("x {}".format(x.shape))
    logging.debug("y {}".format(y.shape))
    logging.debug("sample_weights {}".format(sample_weights.shape))

    output_dir = os.path.join(args.output, "plot_set")
    os.makedirs(output_dir, exist_ok=True)
    subprocess.run("rm -v {}/{}.*.png".format(
        output_dir, config["identifier"]), shell=True)

    for i, channel in enumerate(config['channels']):
        output_path = os.path.join(output_dir, "{}.{:03d}_{}.png".
                                   format(config["identifier"], i, channel))
        logging.info("Producing {}".format(output_path))

        fig, ax = plt.subplots()
        ax.contourf(x[0, ..., i], levels=args.levels)
        plt.savefig(output_path)
        plt.close()

    for i in range(config['n_forecast_days']):
        output_path = os.path.join(output_dir, "{}.y.{:03d}.png".
                                   format(config["identifier"], i + 1))
        y_out = y[0, ..., i, 0]

        logging.info("Producing {}".format(output_path))

        if len(y_out.flatten()) == np.isnan(y_out).sum():
            logging.warning("Skipping {} due to fully nan".format(output_path))
        else:
            fig, ax = plt.subplots()
            ax.contourf(y_out, levels=args.levels)
            plt.savefig(output_path)
            plt.close()



import logging
import subprocess as sp

from enum import Flag, auto
from functools import wraps


class Hemisphere(Flag):
    """

    """

    NONE = 0
    NORTH = auto()
    SOUTH = auto()
    BOTH = NORTH | SOUTH


class HemisphereMixin:
    """

    """

    _hemisphere = Hemisphere.NONE

    @property
    def hemisphere(self):
        return self._hemisphere

    @property
    def hemisphere_str(self):
        return ["north"] if self.north else \
               ["south"] if self.south else \
               ["north", "south"]

    @property
    def hemisphere_loc(self):
        return [90, -180, 0, 180] if self.north else \
               [0, -180, -90, 180] if self.south else \
               [90, -180, -90, 180]

    @property
    def north(self):
        return (self._hemisphere & Hemisphere.NORTH) == Hemisphere.NORTH

    @property
    def south(self):
        return (self._hemisphere & Hemisphere.SOUTH) == Hemisphere.SOUTH

    @property
    def both(self):
        return self._hemisphere & Hemisphere.BOTH


def run_command(command: str, dry: bool = False):
    """Run a shell command

    A wrapper in case we want some additional handling to go in here

    :param command:
    :param dry:
    :return:

    """
    if dry:
        logging.info("Skipping dry commaand: {}".format(command))
        return 0

    ret = sp.run(command, shell=True)
    if ret.returncode < 0:
        logging.warning("Child was terminated by signal: {}".
                        format(-ret.returncode))
    else:
        logging.info("Child returned: {}".format(-ret.returncode))

    return ret


def setup_logging(func,
                  log_format="[%(asctime)-17s :%(levelname)-8s] - %(message)s"):
    @wraps(func)
    def wrapper(*args, **kwargs):
        parsed_args = func(*args, **kwargs)
        level = logging.INFO

        if hasattr(parsed_args, "verbose") and parsed_args.verbose:
            level = logging.DEBUG

        logging.basicConfig(
            level=level,
            format=log_format,
            datefmt="%d-%m-%y %T",
        )

        # TODO: better way of handling these on a case by case basis
        logging.getLogger("cdsapi").setLevel(logging.WARNING)
        logging.getLogger("matplotlib").setLevel(logging.WARNING)
        logging.getLogger("matplotlib.pyplot").setLevel(logging.WARNING)
        logging.getLogger("requests").setLevel(logging.WARNING)
        logging.getLogger("tensorflow").setLevel(logging.WARNING)
        logging.getLogger("urllib3").setLevel(logging.WARNING)
        return parsed_args
    return wrapper




import datetime as dt
import re
import os


def date_arg(string: str) -> object:
    """

    :param string:
    :return:
    """
    d_match = re.search(r'^(\d+)-(\d+)-(\d+)$', string).groups()

    if d_match:
        return dt.date(*[int(s) for s in d_match])


def destination_filename(destination: object,
                         filename: str,
                         date: object) -> object:
    """

    :param destination:
    :param filename:
    :param date:
    :return:
    """
    return os.path.join(destination,
                        "{}.{}{}".format(
                            os.path.splitext(
                                os.path.basename(filename))[0],
                            date.strftime("%Y-%m-%d"),
                            os.path.splitext(
                                os.path.basename(filename))[1],
                        ))
